"""InvenioRDM reader for Commonmeta"""
import httpx
from pydash import py_

from ..utils import (
    normalize_url,
    normalize_doi,
    dict_to_spdx,
    name_to_fos,
    from_inveniordm,
)
from ..base_utils import compact, wrap, presence, sanitize
from ..author_utils import get_authors
from ..date_utils import strip_milliseconds
from ..doi_utils import doi_as_url, doi_from_url
from ..constants import (
    INVENIORDM_TO_CM_TRANSLATIONS,
    COMMONMETA_RELATION_TYPES,
    Commonmeta,
)


def get_inveniordm(pid: str, **kwargs) -> dict:
    """get_inveniordm"""
    if pid is None:
        return {"state": "not_found"}
    url = normalize_url(pid)
    response = httpx.get(url, timeout=10, follow_redirects=True, **kwargs)
    if response.status_code != 200:
        return {"state": "not_found"}
    return response.json()


def read_inveniordm(data: dict, **kwargs) -> Commonmeta:
    """read_inveniordm"""
    meta = data
    read_options = kwargs or {}

    _id = doi_as_url(meta.get("doi", None))
    resource_type = py_.get(meta, "metadata.resource_type.type")
    _type = INVENIORDM_TO_CM_TRANSLATIONS.get(resource_type, "Other")
    contributors = get_authors(
        from_inveniordm(wrap(py_.get(meta, "metadata.creators")))
    )
    # contrib = get_authors(wrap(meta.get("metadata.contributors", None)))
    # if contrib:
    #     contributors = contributors + contrib

    publisher = {"name": meta.get("publisher", None) or "Zenodo"}

    title = py_.get(meta, "metadata.title")
    titles = [{"title": sanitize(title)}] if title else None

    date: dict = {}
    date["published"] = py_.get(meta, ("metadata.publication_date"))
    date["updated"] = strip_milliseconds(meta.get("updated", None))
    container = compact(
        {
            "id": "https://www.re3data.org/repository/r3d100010468",
            "type": "DataRepository" if _type == "Dataset" else "Repository",
            "title": "Zenodo",
        }
    )
    license_ = py_.get(meta, "metadata.license.id")
    if license_:
        license_ = dict_to_spdx({"id": license_})

    descriptions = format_descriptions(
        [
            py_.get(meta, "metadata.description"),
            py_.get(meta, "metadata.notes"),
        ]
    )
    language = py_.get(meta, "metadata.language")
    subjects = [name_to_fos(i) for i in wrap(py_.get(meta, "metadata.keywords"))]

    references = get_references(wrap(py_.get(meta, "metadata.related_identifiers")))
    related_identifiers = get_related_identifiers(
        wrap(py_.get(meta, "metadata.related_identifiers"))
    )
    if meta.get("conceptdoi", None):
        related_identifiers.append(
            {
                "id": doi_as_url(meta.get("conceptdoi")),
                "type": "IsVersionOf",
            }
        )
    files = [get_file(i) for i in wrap(meta.get("files"))]

    state = "findable" if meta or read_options else "not_found"

    return {
        # required properties
        "id": _id,
        "type": _type,
        "doi": doi_from_url(_id),
        "url": normalize_url(py_.get(meta, "links.self_html")),
        "contributors": presence(contributors),
        "titles": titles,
        "publisher": publisher,
        "date": compact(date),
        # recommended and optional properties
        # "additional_type": additional_type,
        "subjects": presence(subjects),
        "language": language,
        # "alternate_identifiers": presence(meta.get("alternateIdentifiers", None)),
        "sizes": None,
        "formats": None,
        "version": py_.get(meta, "metadata.version"),
        "license": presence(license_),
        "descriptions": descriptions,
        "geo_locations": None,
        # "funding_references": presence(meta.get("fundingReferences", None)),
        # "references": presence(references),
        "related_identifiers": presence(related_identifiers),
        # other properties
        "files": files,
        "container": container,
        "provider": "InvenioRDM",
        "state": state,
        # "schema_version": meta.get("schemaVersion", None),
    } | read_options


def get_references(references: list) -> list:
    """get_references"""

    def is_reference(reference):
        """is_reference"""
        return reference.get("relationType", None) in ["Cites", "References"]

    def map_reference(reference):
        """map_reference"""
        identifier = reference.get("relatedIdentifier", None)
        identifier_type = reference.get("relatedIdentifierType", None)
        if identifier and identifier_type == "DOI":
            reference["doi"] = normalize_doi(identifier)
        elif identifier and identifier_type == "URL":
            reference["url"] = normalize_url(identifier)
        reference = py_.omit(
            reference,
            [
                "relationType",
                "relatedIdentifier",
                "relatedIdentifierType",
                "resourceTypeGeneral",
                "schemeType",
                "schemeUri",
                "relatedMetadataScheme",
            ],
        )
        return reference

    return [map_reference(i) for i in references if is_reference(i)]


def get_file(file: dict) -> str:
    """get_file"""
    _type = file.get("type", None)
    return compact(
        {
            "bucket": file.get("bucket", None),
            "key": file.get("key", None),
            "checksum": file.get("checksum", None),
            "url": py_.get(file, "links.self"),
            "size": file.get("size", None),
            "mimeType": "application/" + _type if _type else None,
        }
    )


def get_related_identifiers(related_identifiers: list) -> list:
    """get_related_identifiers"""

    def map_related_identifier(related_identifier: dict) -> dict:
        """get_related_identifier"""
        identifier = related_identifier.get("identifier", None)
        scheme = related_identifier.get("scheme", None)
        relation_type = related_identifier.get("relation", None)
        if scheme == "doi":
            identifier = doi_as_url(identifier)
        else:
            identifier = normalize_url(identifier)
        return {
            "id": identifier,
            "type": py_.capitalize(relation_type, False) if relation_type else None,
        }

    identifiers = [map_related_identifier(i) for i in related_identifiers]
    return [
        i
        for i in identifiers
        if py_.upper_first(i["type"]) in COMMONMETA_RELATION_TYPES
    ]


def format_descriptions(descriptions: list) -> list:
    """format_descriptions"""
    return [
        {
            "description": sanitize(i),
            "descriptionType": "Abstract" if index == 0 else "Other",
        }
        for index, i in enumerate(descriptions)
        if i
    ]



"""Bibtex writer for commonmeta-py"""
from bibtexparser.bwriter import BibTexWriter
from bibtexparser.bibdatabase import BibDatabase
from bibtexparser.customization import page_double_hyphen

from ..utils import pages_as_string
from ..base_utils import compact
from ..author_utils import authors_as_string
from ..date_utils import get_month_from_date, get_iso8601_date, MONTH_SHORT_NAMES
from ..doi_utils import doi_from_url
from ..constants import CM_TO_BIB_TRANSLATIONS, Commonmeta


def write_bibtex(metadata: Commonmeta) -> str:
    """Write bibtex"""
    if metadata.write_errors is not None:
        return None
    item = write_bibtex_item(metadata)
    bibtex_str = """
    @comment{
        BibTeX entry created by commonmeta-py
    }
    """
    bib_database = BibDatabase()
    bib_database.entries = [item]
    bib_database.entries[0] = page_double_hyphen(bib_database.entries[0])
    writer = BibTexWriter()
    writer.common_strings = True
    writer.indent = "    "
    bibtex_str = writer.write(bib_database)

    # Hack to remove curly braces around month names
    for month_name in MONTH_SHORT_NAMES:
        bibtex_str = bibtex_str.replace(f"{{{month_name}}}", month_name)
    return bibtex_str


def write_bibtex_item(metadata: Commonmeta) -> dict:
    """Write bibtex item"""
    container = metadata.container if metadata.container else {}
    date_published = get_iso8601_date(metadata.date.get("published", None))
    authors = authors_as_string(metadata.contributors)
    if metadata.titles and len(metadata.titles) > 1:
        title = ": ".join(
            [
                metadata.titles[0].get("title", None),
                metadata.titles[1].get("title", None),
            ]
        )
    elif metadata.titles and len(metadata.titles) == 1:
        title = metadata.titles[0].get("title", None)
    else:
        title = None
    doi = doi_from_url(metadata.id)
    _id = doi if doi else metadata.id
    _type = CM_TO_BIB_TRANSLATIONS.get(metadata.type, "misc")
    abstract = (
        metadata.descriptions[0].get("description", None)
        if metadata.descriptions
        else None
    )
    author = authors if authors and len(authors) > 0 else None
    license_ = str(metadata.license.get("url")) if metadata.license else None
    institution = metadata.publisher.get("name", None) if _type == "phdthesis" else None
    issn = (
        container.get("identifier", None)
        if container.get("identifierType", None) == "ISSN"
        else None
    )
    isbn = (
        container.get("identifier", None)
        if container.get("identifierType", None) == "ISBN"
        else None
    )
    issue = container.get("issue", None)
    journal = (
        container.get("title", None)
        if _type not in ["inbook", "inproceedings"]
        and container.get("type") in ["Journal", "Periodical"]
        else None
    )
    booktitle = (
        container.get("title", None) if _type in ["inbook", "inproceedings"] else None
    )
    language = metadata.language
    location = (
        container.get("location", None)
        if _type not in ["article", "phdthesis"]
        else None
    )
    month = get_month_from_date(date_published)
    pages = pages_as_string(container)
    publisher = (
        metadata.publisher.get("name", None)
        if _type not in ["article", "phdthesis"]
        else None
    )
    series = container.get("series", None)
    url = metadata.url
    year = date_published[:4] if date_published else None

    return compact(
        {
            "ID": _id,
            "ENTRYTYPE": _type,
            "abstract": abstract,
            "author": author,
            "copyright": license_,
            "doi": doi,
            "institution": institution,
            "isbn": isbn,
            "issn": issn,
            "issue": issue,
            "journal": journal,
            "booktitle": booktitle,
            "language": language,
            "location": location,
            "month": month,
            "pages": pages,
            "publisher": publisher,
            "series": series,
            "title": title,
            "url": url,
            "urldate": date_published,
            "year": year,
        }
    )


def write_bibtex_list(metalist):
    """Write bibtex list"""
    if metalist is None:
        return None

    bib_database = BibDatabase()
    bib_database.entries = [write_bibtex_item(item) for item in metalist.items]

    # TODO: Fix page_double_hyphen in write_bibtex_item
    bib_database.entries = [page_double_hyphen(entry) for entry in bib_database.entries]
    writer = BibTexWriter()
    writer.common_strings = True
    writer.indent = "    "
    bibtex_str = writer.write(bib_database)

    # Hack to remove curly braces around month names
    # TODO: Fix this in write_bibtex_item
    for month_name in MONTH_SHORT_NAMES:
        bibtex_str = bibtex_str.replace(f"{{{month_name}}}", month_name)
    return bibtex_str



"""Doi utils for commonmeta-py"""
import re
from typing import Optional
import httpx
from furl import furl

from .base_utils import compact


def validate_doi(doi: Optional[str]) -> Optional[str]:
    """Validate a DOI"""
    if doi is None:
        return None
    match = re.search(
        r"\A(?:(http|https):/(/)?(dx\.)?(doi\.org|handle\.stage\.datacite\.org|handle\.test\.datacite\.org)/)?(doi:)?(10\.\d{4,5}/.+)\Z",  # noqa: E501
        doi,
    )
    if match is None:
        return None
    return match.group(6)


def validate_prefix(doi: Optional[str]) -> Optional[str]:
    """Validate a DOI prefix for a given DOI"""
    if doi is None:
        return None
    match = re.search(
        r"\A(?:(http|https):/(/)?(dx\.)?(doi\.org|handle\.stage\.datacite\.org|handle\.test\.datacite\.org)/)?(doi:)?(10\.\d{4,5}).*\Z",  # noqa: E501
        doi,
    )
    if match is None:
        return None
    return match.group(6)


def doi_from_url(url: Optional[str]) -> Optional[str]:
    """Return a DOI from a URL"""
    if url is None:
        return None
    match = re.search(
        r"\A(?:(http|https)://(dx\.)?(doi\.org|handle\.stage\.datacite\.org|handle\.test\.datacite\.org)/)?(doi:)?(10\.\d{4,5}/.+)\Z",
        url,
    )
    if match is None:
        return None
    return match.group(5).lower()


def doi_as_url(doi: Optional[str]) -> Optional[str]:
    """Return a DOI as a URL"""
    if doi is None:
        return None
    return "https://doi.org/" + doi.lower()


def normalize_doi(doi: Optional[str], **kwargs) -> Optional[str]:
    """Normalize a DOI"""
    doi_str = validate_doi(doi)
    if not doi_str:
        return None
    return doi_resolver(doi, **kwargs) + doi_str.lower()


def doi_resolver(doi, **kwargs):
    """Return a DOI resolver for a given DOI"""
    if doi is None:
        return None
    match = re.match(
        r"\A(http|https):/(/)?handle\.stage\.datacite\.org", doi, re.IGNORECASE
    )
    if match is not None or kwargs.get("sandbox", False):
        return "https://handle.stage.datacite.org/"
    return "https://doi.org/"


def get_doi_ra(doi) -> Optional[str]:
    """Return the DOI registration agency for a given DOI"""
    prefix = validate_prefix(doi)
    if prefix is None:
        return None
    response = httpx.get("https://doi.org/ra/" + prefix, timeout=10)
    if response.status_code != 200:
        return None
    return response.json()[0].get("RA", None)


def get_crossref_member(member_id) -> Optional[dict]:
    """Return the Crossref member for a given member_id"""
    response = httpx.get("https://api.crossref.org/members/" + member_id, timeout=10)
    if response.status_code != 200:
        return None
    data = response.json().get("message", None)
    name = data.get("primary-name", None)
    return {"id": "https://api.crossref.org/members/" + member_id, "name": name}


def crossref_api_url(doi: str) -> str:
    """Return the Crossref API URL for a given DOI"""
    return "https://api.crossref.org/works/" + doi


def crossref_xml_api_url(doi: str) -> str:
    """Return the Crossref XML API URL for a given DOI"""
    return f"https://api.crossref.org/works/{doi}/transform/application/vnd.crossref.unixsd+xml"


def crossref_api_query_url(query: dict) -> str:
    """Return the Crossref API query URL"""
    url = "https://api.crossref.org/works"
    f = furl(url)
    rows = min(int(query.get("rows", 20)), 1000)
    queries = []
    filters = []
    if query.get("query", None) is not None:
        queries += [query.get("query")]
    for key, value in query.items():
        if key in [
            "query.bibliographic",
            "query.author",
            "query.title",
            "query.container-title",
        ]:
            queries += [f"{key}:{value}"]
        _query = ",".join(queries) if len(queries) > 0 else None

    for key, value in query.items():
        if key in [
            "prefix",
            "member",
            "type",
            "has-full-text",
            "has-references",
            "has-orcid",
            "has-funder",
            "has-license",
        ]:
            filters += [f"{key}:{value}"]
        _filter = ",".join(filters) if len(filters) > 0 else None
    f.args = compact({"rows": rows, "query": _query, "filter": _filter})

    return f.url


def crossref_api_sample_url(number: int = 1, **kwargs) -> str:
    """Return the Crossref API URL for a sample of works"""
    types = [
        "book-section",
        "monograph",
        "report-component",
        "report",
        "peer-review",
        "book-track",
        "journal-article",
        "book-part",
        "other",
        "book",
        "journal-volume",
        "book-set",
        "reference-entry",
        "proceedings-article",
        "journal",
        "component",
        "book-chapter",
        "proceedings-series",
        "report-series",
        "proceedings",
        "database",
        "standard",
        "reference-book",
        "posted-content",
        "journal-issue",
        "dissertation",
        "grant",
        "dataset",
        "book-series",
        "edited-book",
        "journal-section",
        "monograph-series",
        "journal-meta",
        "book-series-meta",
        "component-list",
        "journal-issue-meta",
        "journal-meta",
        "book-part-meta",
        "book-meta",
        "proceedings-meta",
        "book-series-meta",
        "book-set",
    ]
    url = f"https://api.crossref.org/works?sample={number}"
    if kwargs.get("prefix", None) and validate_prefix(kwargs.get("prefix")):
        url += f"&filter=prefix:{kwargs.get('prefix')}"
    if kwargs.get("_type", None) and kwargs.get("_type") in types:
        url += f"&filter=type:{kwargs.get('_type')}"
    return url


def datacite_api_url(doi: str, **kwargs) -> str:
    """Return the DataCite API URL for a given DOI"""
    match = re.match(
        r"\A(http|https):/(/)?handle\.stage\.datacite\.org", doi, re.IGNORECASE
    )
    if match is not None or kwargs.get("sandbox", False):
        return f"https://api.stage.datacite.org/dois/{doi_from_url(doi)}?include=media,client"
    return f"https://api.datacite.org/dois/{doi_from_url(doi)}?include=media,client"


def datacite_api_sample_url(number: int = 1, **kwargs) -> str:
    """Return the DataCite API URL for a sample of dois"""
    if kwargs.get("sandbox", False):
        return f"https://api.stage.datacite.org/dois?random=true&page[size]={number}"
    return f"https://api.datacite.org/dois?random=true&page[size]={number}"


def is_rogue_scholar_doi(doi: str) -> bool:
    """Return True if DOI is from Rogue Scholar"""
    prefix = validate_prefix(doi)
    return prefix in [
        "10.34732",
        "10.53731",
        "10.54900",
        "10.57689",
        "10.59348",
        "10.59349",
        "10.59350",
    ]



"""RIS writer for commonmeta-py"""
from ..utils import to_ris
from ..base_utils import compact, wrap, presence, parse_attributes
from ..doi_utils import doi_from_url
from ..constants import CM_TO_RIS_TRANSLATIONS


def write_ris(metadata):
    """Write ris"""
    container = metadata.container or {}
    _type = CM_TO_RIS_TRANSLATIONS.get(metadata.type, "GEN")
    ris = compact(
        {
            "TY": _type,
            "T1": parse_attributes(metadata.titles, content="title", first=True),
            "T2": container.get("title", None),
            "AU": to_ris(metadata.contributors),
            "DO": doi_from_url(metadata.id),
            "UR": metadata.url,
            "AB": parse_attributes(
                metadata.descriptions, content="description", first=True
            ),
            "KW": presence(
                parse_attributes(
                    wrap(metadata.subjects), content="subject", first=False
                )
            ),
            "PY": metadata.date.get("published")[:4]
            if metadata.date.get("published", None)
            else None,
            "PB": metadata.publisher.get("name", None),
            "LA": metadata.language,
            "VL": container.get("volume", None),
            "IS": container.get("issue", None),
            "SP": container.get("firstPage", None),
            "EP": container.get("lastPage", None),
            # 'SN'= > Array.wrap(related_identifiers).find do | ri |
            # ri['relationType'] == 'IsPartOf'
            # end.to_h.fetch('relatedIdentifier', nil),
            "ER": "",
        }
    )
    string = []
    for key, val in ris.items():
        if isinstance(val, list) and val not in [[], [None]]:
            for vai in val:
                string.append(f"{key}  - {vai}")
        elif val not in [[], [None]]:
            string.append(f"{key}  - {val}")
    return "\r\n".join(string)


def write_ris_list(metalist):
    """Write RIS list"""
    if metalist is None:
        return None
    items = [write_ris(item) for item in metalist.items]
    return "\r\n\r\n".join(items)



"""Author utils module for commonmeta-py"""
import re
from typing import List
from nameparser import HumanName
from pydash import py_

from .utils import (
    normalize_orcid,
    normalize_id,
    normalize_ror,
    normalize_isni,
    format_name_identifier,
    validate_ror,
    validate_orcid,
)
from .base_utils import parse_attributes, wrap, presence, compact

from .constants import (
    COMMONMETA_CONTRIBUTOR_ROLES,
)


def get_one_author(author, **kwargs):
    """parse one author string into commonmeta format"""
    # if author is a string
    if isinstance(author, str):
        author = {"creatorName": author}

    # malformed XML
    if isinstance(author.get("creatorName", None), list):
        return None

    name = (
        parse_attributes(author.get("creatorName", None))
        or parse_attributes(author.get("contributorName", None))
        or parse_attributes(author.get("name", None))
    )
    given_name = parse_attributes(author.get("givenName", None)) or parse_attributes(
        author.get("given", None)
    )
    family_name = parse_attributes(author.get("familyName", None)) or parse_attributes(
        author.get("family", None)
    )

    name = cleanup_author(name)

    # make sure we have a name
    if not name and not given_name and not family_name:
        return None

    # parse contributor roles, checking for roles supported by commonmeta
    contributor_roles = wrap(
        parse_attributes(author.get("contributorType", None))
    ) or wrap(parse_attributes(author.get("contributor_roles", None)))
    contributor_roles = [
        i for i in contributor_roles if i in COMMONMETA_CONTRIBUTOR_ROLES
    ] or ["Author"]

    # parse author type, i.e. "Person", "Organization" or not specified
    _type = parse_attributes(
        author.get("creatorName", None), content="type", first=True
    ) or parse_attributes(
        author.get("contributorName", None), content="type", first=True
    )

    # also handle Crossref, JSON Feed, or DataCite metadata
    _id = (
        author.get("id", None)
        or author.get("ORCID", None)
        or author.get("url", None)
        or next(
            (
                format_name_identifier(i)
                for i in wrap(author.get("nameIdentifiers", None))
            ),
            None,
        )
    )
    _id = normalize_orcid(_id) or normalize_ror(_id) or normalize_isni(_id) or _id

    # DataCite metadata
    if isinstance(_type, str) and _type.endswith("al"):
        _type = _type[:-3]

    if not _type and isinstance(_id, str) and validate_ror(_id) is not None:
        _type = "Organization"
    elif not _type and isinstance(_id, str) and validate_orcid(_id) is not None:
        _type = "Person"
    elif not _type and (given_name or family_name):
        _type = "Person"
    elif not _type and name and kwargs.get("via", None) == "crossref":
        _type = "Organization"
    elif not _type and is_personal_name(name):
        _type = "Person"
    elif not _type and name:
        _type = "Organization"

    # split name for type Person into given/family name if not already provided
    if _type == "Person" and name and not given_name and not family_name:
        names = HumanName(name)

        if names:
            given_name = (
                " ".join([names.first, names.middle]).strip() if names.first else None
            )
            family_name = names.last if names.last else None
        else:
            given_name = None
            family_name = None

    # return author in commonmeta format, using name vs. given/family name
    # depending on type
    return compact(
        {
            "id": _id,
            "type": _type,
            "contributorRoles": contributor_roles,
            "name": name if _type == "Organization" else None,
            "givenName": given_name if _type == "Person" else None,
            "familyName": family_name if _type == "Person" else None,
            "affiliation": presence(
                get_affiliations(wrap(author.get("affiliation", None)))
            ),
        }
    )


def is_personal_name(name):
    """is_personal_name"""
    # personal names are not allowed to contain semicolons
    if ";" in name:
        return False

    # check if a name has only one word, e.g. "FamousOrganization", not including commas
    if len(name.split(" ")) == 1 and "," not in name:
        return False

    # check if name contains words known to be used in organization names
    if any(
        word in name
        for word in [
            "University",
            "College",
            "Institute",
            "School",
            "Center",
            "Department",
            "Laboratory",
            "Library",
            "Museum",
            "Foundation",
            "Society",
            "Association",
            "Company",
            "Corporation",
            "Collaboration",
            "Consortium",
            "Incorporated",
            "Inc.",
            "Institut",
            "Research",
            "Science",
            "Team",
            "Ministry",
            "Government",
        ]
    ):
        return False

    # check for suffixes, e.g. "John Smith, MD"
    if name.split(", ")[-1] in ["MD", "PhD", "BS"]:
        return True

    # check of name can be parsed into given/family name
    names = HumanName(name)
    if names and (names.first or names.last):
        return True

    return False


def cleanup_author(author):
    """clean up author string"""
    if author is None:
        return None
    
    if author.startswith(","):
        return None

    # detect pattern "Smith J.", but not "Smith, John K."
    if "," not in author:
        author = re.sub(r"/([A-Z]\.)?(-?[A-Z]\.)/", ", \1\2", author)

    # remove spaces around hyphens
    author = author.replace(" - ", "-")

    # remove non-standard space characters
    author = re.sub("/[ \t\r\n\v\f]/", " ", author)
    return author


def get_authors(authors, **kwargs):
    """transform array of author dicts into commonmeta format"""
    return py_.uniq(py_.compact([get_one_author(i, **kwargs) for i in authors]))


def authors_as_string(authors: List[dict]) -> str:
    """convert authors list to string, e.g. for bibtex"""

    def format_author(author):
        if author.get("familyName", None) and author.get("givenName", None):
            return f"{author['familyName']}, {author['givenName']}"
        elif author.get("familyName", None):
            return author["familyName"]
        return author.get("name", None)

    return " and ".join([format_author(i) for i in wrap(authors) if i is not None])


def get_affiliations(affiliations: List[dict]) -> List[dict]:
    """parse array of affiliation strings into commonmeta format"""

    def format_element(i):
        """format single affiliation element"""
        affiliation_identifier = None
        if isinstance(i, str):
            name = i
            scheme_uri = None
        else:
            if i.get("affiliationIdentifier", None) is not None:
                affiliation_identifier = i["affiliationIdentifier"]
                if i.get("schemeURI", None) is not None:
                    scheme_uri = (
                        i["schemeURI"]
                        if i["schemeURI"].endswith("/")
                        else "{affiliation['schemeURI']}/"
                    )
                affiliation_identifier = (
                    normalize_id(scheme_uri + affiliation_identifier)
                    if (
                        not affiliation_identifier.startswith("https://")
                        and scheme_uri is not None
                    )
                    else normalize_id(affiliation_identifier)
                )
            name = i.get("name", None) or i.get("#text", None)
        return compact(
            {
                "id": affiliation_identifier,
                "name": name,
            }
        )

    return py_.uniq(py_.compact([format_element(i) for i in affiliations]))



"""crossref reader for commonmeta-py"""
from typing import Optional
import httpx
from pydash import py_

from ..utils import (
    dict_to_spdx,
    normalize_cc_url,
    normalize_url,
    normalize_doi,
    normalize_issn,
)
from ..base_utils import wrap, compact, presence, sanitize, parse_attributes
from ..author_utils import get_authors
from ..date_utils import get_date_from_date_parts
from ..doi_utils import (
    doi_as_url,
    doi_from_url,
    # get_doi_ra,
    # get_crossref_member,
    crossref_api_url,
    crossref_api_query_url,
    crossref_api_sample_url,
)
from ..constants import (
    CR_TO_CM_TRANSLATIONS,
    CR_TO_CM_CONTAINER_TRANSLATIONS,
    CROSSREF_CONTAINER_TYPES,
    Commonmeta,
)


def get_crossref_list(query: dict, **kwargs) -> list[dict]:
    """get_crossref list from Crossref API."""
    url = crossref_api_query_url(query, **kwargs)
    response = httpx.get(url, timeout=30, **kwargs)
    if response.status_code != 200:
        return []
    return response.json().get("message", {}).get("items", [])


def get_crossref(pid: str, **kwargs) -> dict:
    """get_crossref"""
    doi = doi_from_url(pid)
    if doi is None:
        return {"state": "not_found"}
    url = crossref_api_url(doi)
    response = httpx.get(url, timeout=10, **kwargs)
    if response.status_code != 200:
        return {"state": "not_found"}
    return response.json().get("message", {})


def read_crossref(data: Optional[dict], **kwargs) -> Commonmeta:
    """read_crossref"""
    if data is None:
        return {"state": "not_found"}
    meta = data
    # read_options = ActiveSupport::HashWithIndifferentAccess.
    # new(options.except(:doi, :id, :url,
    # :sandbox, :validate, :ra))
    read_options = kwargs or {}

    doi = meta.get("DOI", None)
    _id = doi_as_url(doi)
    _type = CR_TO_CM_TRANSLATIONS.get(meta.get("type", None)) or "Other"

    if meta.get("author", None):
        contributors = get_authors(wrap(meta.get("author")), via="crossref")
    else:
        contributors = []

    def editor_type(item):
        item["contributorType"] = "Editor"
        return item

    editors = [editor_type(i) for i in wrap(meta.get("editor", None))]
    if editors:
        contributors += get_authors(editors)

    url = normalize_url(py_.get(meta, "resource.primary.URL"))
    titles = get_titles(meta)
    publisher = compact({"name": meta.get("publisher", None)})

    date = compact(
        {
            "published": py_.get(meta, "issued.date-time")
            or get_date_from_date_parts(meta.get("issued", None))
            or py_.get(meta, "created.date-time")
        }
    )

    license_ = meta.get("license", None)
    if license_ is not None:
        license_ = normalize_cc_url(license_[0].get("URL", None))
        license_ = dict_to_spdx({"url": license_}) if license_ else None

    container = get_container(meta)
    references =py_.uniq([get_reference(i) for i in wrap(meta.get("reference", None))])
    funding_references = from_crossref_funding(wrap(meta.get("funder", None)))

    description = meta.get("abstract", None)
    if description is not None:
        descriptions = [
            {"description": sanitize(description), "descriptionType": "Abstract"}
        ]
    else:
        descriptions = None

    subjects = py_.uniq([{"subject": i} for i in wrap(meta.get("subject", []))])
    files = [
        get_file(i)
        for i in wrap(meta.get("link", None))
        if i["content-type"] != "unspecified"
    ]

    state = "findable" if meta or read_options else "not_found"

    return {
        # required properties
        "id": _id,
        "type": _type,
        "url": url,
        "contributors": presence(contributors),
        "titles": presence(titles),
        "publisher": presence(publisher),
        "date": presence(date),
        # recommended and optional properties
        "subjects": presence(subjects),
        "language": meta.get("language", None),
        "alternate_identifiers": None,
        "sizes": None,
        "formats": None,
        "version": meta.get("version", None),
        "license": license_,
        "descriptions": descriptions,
        "geo_locations": None,
        "funding_references": presence(funding_references),
        "references": presence(references),
        # other properties
        "files": presence(files),
        "container": presence(container),
        "provider": "Crossref",  # get_doi_ra(_id),
        "state": state,
        "schema_version": None,
    } | read_options


def get_titles(meta):
    """Title information from Crossref metadata."""
    titles = wrap(parse_attributes(meta.get("title", None)))
    subtitles = wrap(parse_attributes(meta.get("subtitle", None)))
    original_language_titles = wrap(
        parse_attributes(meta.get("original_language_title", None))
    )
    language = None
    return (
        [{"title": sanitize(i)} for i in titles]
        + [
            compact(
                {
                    "title": sanitize(i),
                    "titleType": "Subtitle",
                }
            )
            for i in subtitles
        ]
        + [
            compact(
                {
                    "title": sanitize(i),
                    "titleType": "TranslatedTitle",
                    "lang": language,
                }
            )
            for i in original_language_titles
        ]
    )


def get_reference(reference: Optional[dict]) -> Optional[dict]:
    """Get reference from Crossref reference"""
    if reference is None or not isinstance(reference, dict):
        return None
    doi = reference.get("DOI", None)
    metadata = {
        "key": reference.get("key", None),
        "doi": normalize_doi(doi) if doi else None,
        "contributor": reference.get("author", None),
        "title": reference.get("article-title", None),
        "publisher": reference.get("publisher", None),
        "publicationYear": reference.get("year", None),
        "volume": reference.get("volume", None),
        "issue": reference.get("issue", None),
        "firstPage": reference.get("first-page", None),
        "lastPage": reference.get("last-page", None),
        "containerTitle": reference.get("journal-title", None),
        "edition": None,
        "unstructured": reference.get("unstructured", None) if doi is None else None,
    }
    return compact(metadata)


def get_file(file: dict) -> dict:
    """Get file from Crossref"""
    return compact(
        {
            "url": file.get("URL", None),
            "mimeType": file.get("content-type", None),
        }
    )


def get_container(meta: dict) -> dict:
    """Get container from Crossref"""
    container_type = CROSSREF_CONTAINER_TYPES.get(meta.get("type", None))
    container_type = CR_TO_CM_CONTAINER_TRANSLATIONS.get(container_type, None)
    issn = (
        next(
            (
                item
                for item in wrap(meta.get("issn-type", None))
                if item["type"] == "electronic"
            ),
            None,
        )
        or next(
            (
                item
                for item in wrap(meta.get("issn-type", None))
                if item["type"] == "print"
            ),
            None,
        )
        or {}
    )
    issn = normalize_issn(issn["value"]) if issn else None
    isbn = (
        next(
            (
                item
                for item in wrap(meta.get("isbn-type", None))
                if item["type"] == "electronic"
            ),
            None,
        )
        or next(
            (
                item
                for item in wrap(meta.get("isbn-type", None))
                if item["type"] == "print"
            ),
            None,
        )
        or {}
    )
    isbn = isbn["value"] if isbn else None
    container_title = parse_attributes(meta.get("container-title", None), first=True)
    volume = meta.get("volume", None)
    issue = py_.get(meta, "journal-issue.issue")
    if meta.get("page", None):
        pages = meta.get("page", None).split("-")
        first_page = pages[0]
        last_page = pages[1] if len(pages) > 1 else None
    else:
        first_page = None
        last_page = None

    # TODO: add support for series, location, missing in Crossref JSON

    return compact(
        {
            "type": container_type,
            "identifier": issn or isbn,
            "identifierType": "ISSN" if issn else "ISBN" if isbn else None,
            "title": container_title,
            "volume": volume,
            "issue": issue,
            "firstPage": first_page,
            "lastPage": last_page,
        }
    )


def from_crossref_funding(funding_references: list) -> list:
    """Get funding references from Crossref"""
    formatted_funding_references = []
    for funding in funding_references:
        funding_reference = compact(
            {
                "funderName": funding.get("name", None),
                "funderIdentifier": doi_as_url(funding["DOI"])
                if funding.get("DOI", None) is not None
                else None,
                "funderIdentifierType": "Crossref Funder ID"
                if funding.get("DOI", "").startswith("10.13039")
                else None,
            }
        )
        funding_reference = py_.omit(funding_reference, "DOI", "doi-asserted-by")
        if (
            funding.get("name", None) is not None
            and funding.get("award", None) is not None
        ):
            for award in wrap(funding["award"]):
                fund_ref = funding_reference.copy()
                fund_ref["awardNumber"] = award
                formatted_funding_references.append(fund_ref)
        elif funding_reference != {}:
            formatted_funding_references.append(funding_reference)
    return formatted_funding_references


def get_random_crossref_id(number: int = 1, **kwargs) -> list:
    """Get random DOI from Crossref"""
    number = 20 if number > 20 else number
    url = crossref_api_sample_url(number, **kwargs)
    try:
        response = httpx.get(url, timeout=10)
        if response.status_code != 200:
            return []

        items = py_.get(response.json(), "message.items")
        return [i.get("DOI") for i in items]
    except (httpx.ReadTimeout, httpx.ConnectError):
        return []



"""Constants for commonmeta-py"""
from typing import Optional, TypedDict, List


class Commonmeta(TypedDict):
    """TypedDict for Commonmeta"""

    id: str
    type: str
    url: str
    creators: List[dict]
    titles: List[dict]
    publisher: dict
    date: dict
    additional_type: Optional[str]
    subjects: Optional[List[dict]]
    contributors: Optional[List[dict]]
    language: Optional[str]
    alternate_identifiers: Optional[List[dict]]
    related_identifiers: Optional[List[dict]]
    sizes: Optional[List[dict]]
    formats: Optional[List[dict]]
    version: Optional[str]
    license: Optional[dict]
    descriptions: Optional[List[dict]]
    geo_locations: Optional[List[dict]]
    funding_references: Optional[List[dict]]
    references: Optional[List[dict]]
    container: Optional[dict]
    files: Optional[List[dict]]
    agency: Optional[str]
    state: str
    schema_version: Optional[str]


# source: https://www.bibtex.com/e/entry-types/
BIB_TO_CM_TRANSLATIONS = {
    "article": "JournalArticle",
    "book": "Book",
    "booklet": "Book",
    "inbook": "BookChapter",
    "inproceedings": "ProceedingsArticle",
    "manual": "Report",
    "mastersthesis": "Dissertation",
    "misc": "Other",
    "phdthesis": "Dissertation",
    "proceedings": "Proceedings",
    "techreport": "Report",
    "unpublished": "Manuscript",
}

CM_TO_BIB_TRANSLATIONS = {
    "Article": "article",
    "Book": "book",
    "BookChapter": "inbook",
    "Dissertation": "phdthesis",
    "JournalArticle": "article",
    "Manuscript": "unpublished",
    "Other": "misc",
    "Proceedings": "proceedings",
    "ProceedingsArticle": "inproceedings",
    "Report": "techreport",
}

# source: https://docs.citationstyles.org/en/stable/specification.html?highlight=book#appendix-iii-types
CSL_TO_CM_TRANSLATIONS = {
    "article": "Article",
    "article-journal": "JournalArticle",
    "article-magazine": "Article",
    "article-newspaper": "Article",
    "bill": "LegalDocument",
    "book": "Book",
    "broadcast": "Audiovisual",
    "chapter": "BookChapter",
    "classic": "Book",
    "collection": "Collection",
    "dataset": "Dataset",
    "document": "Document",
    "entry": "Entry",
    "entry-dictionary": "Entry",
    "entry-encyclopedia": "Entry",
    "event": "Event",
    "figure": "Figure",
    "graphic": "Image",
    "hearing": "LegalDocument",
    "interview": "Document",
    "legal_case": "LegalDocument",
    "legislation": "LegalDocument",
    "manuscript": "Manuscript",
    "map": "Map",
    "motion_picture": "Audiovisual",
    "musical_score": "Document",
    "pamphlet": "Document",
    "paper-conference": "ProceedingsArticle",
    "patent": "Patent",
    "performance": "Performance",
    "periodical": "Journal",
    "personal_communication": "PersonalCommunication",
    "post": "Post",
    "post-weblog": "Article",
    "regulation": "LegalDocument",
    "report": "Report",
    "review": "Review",
    "review-book": "Review",
    "software": "Software",
    "song": "Audiovisual",
    "speech": "Presentation",
    "standard": "Standard",
    "thesis": "Dissertation",
    "treaty": "LegalDocument",
    "webpage": "WebPage",
}

CM_TO_CSL_TRANSLATIONS = {
    "Article": "article",
    "JournalArticle": "article-journal",
    "Book": "book",
    "BookChapter": "chapter",
    "Collection": "collection",
    "Dataset": "dataset",
    "Document": "document",
    "Entry": "entry",
    "Event": "event",
    "Figure": "figure",
    "Image": "graphic",
    "LegalDocument": "legal_case",
    "Manuscript": "manuscript",
    "Map": "map",
    "Audiovisual": "motion_picture",
    "Patent": "patent",
    "Performance": "performance",
    "Journal": "periodical",
    "PersonalCommunication": "personal_communication",
    "Report": "report",
    "Review": "review",
    "Software": "software",
    "Presentation": "speech",
    "Standard": "standard",
    "Dissertation": "thesis",
    "WebPage": "webpage",
}

# source: http://api.crossref.org/types
CR_TO_CM_TRANSLATIONS = {
    "book-chapter": "BookChapter",
    "book-part": "BookPart",
    "book-section": "BookSection",
    "book-series": "BookSeries",
    "book-set": "BookSet",
    "book-track": "BookTrack",
    "book": "Book",
    "component": "Component",
    "database": "Database",
    "dataset": "Dataset",
    "dissertation": "Dissertation",
    "edited-book": "Book",
    "grant": "Grant",
    "journal-article": "JournalArticle",
    "journal-issue": "JournalIssue",
    "journal-volume": "JournalVolume",
    "journal": "Journal",
    "monograph": "Book",
    "other": "Other",
    "peer-review": "PeerReview",
    "posted-content": "Article",
    "proceedings-article": "ProceedingsArticle",
    "proceedings-series": "ProceedingsSeries",
    "proceedings": "Proceedings",
    "reference-book": "Book",
    "reference-entry": "Entry",
    "report-component": "ReportComponent",
    "report-series": "ReportSeries",
    "report": "Report",
    "standard": "Standard",
}

CM_TO_CR_TRANSLATIONS = {
    "Article": "PostedContent",
    "BookChapter": "BookChapter",
    "BookSeries": "BookSeries",
    "Book": "Book",
    "Component": "Component",
    "Dataset": "Dataset",
    "Dissertation": "Dissertation",
    "Grant": "Grant",
    "JournalArticle": "JournalArticle",
    "JournalIssue": "JournalIssue",
    "JournalVolume": "JournalVolume",
    "Journal": "Journal",
    "ProceedingsArticle": "ProceedingsArticle",
    "ProceedingsSeries": "ProceedingsSeries",
    "Proceedings": "Proceedings",
    "ReportComponent": "ReportComponent",
    "ReportSeries": "ReportSeries",
    "Report": "Report",
    "Review": "PeerReview",
    "Other": "Other",
}

# source: https://github.com/datacite/schema/blob/master/source/meta/kernel-4/include/datacite-resourceType-v4.xsd
DC_TO_CM_TRANSLATIONS = {
    "Audiovisual": "Audiovisual",
    "BlogPosting": "Article",
    "Book": "Book",
    "BookChapter": "BookChapter",
    "Collection": "Collection",
    "ComputationalNotebook": "ComputationalNotebook",
    "ConferencePaper": "ProceedingsArticle",
    "ConferenceProceeding": "Proceedings",
    "DataPaper": "JournalArticle",
    "Dataset": "Dataset",
    "Dissertation": "Dissertation",
    "Event": "Event",
    "Image": "Image",
    "Instrument": "Instrument",
    "InteractiveResource": "InteractiveResource",
    "Journal": "Journal",
    "JournalArticle": "JournalArticle",
    "Model": "Model",
    "OutputManagementPlan": "OutputManagementPlan",
    "PeerReview": "PeerReview",
    "PhysicalObject": "PhysicalObject",
    "Poster": "Presentation",
    "Preprint": "Article",
    "Report": "Report",
    "Service": "Service",
    "Software": "Software",
    "Sound": "Sound",
    "Standard": "Standard",
    "StudyRegistration": "StudyRegistration",
    "Text": "Document",
    "Thesis": "Dissertation",
    "Workflow": "Workflow",
    "Other": "Other",
}

# https://github.com/zenodo/zenodo/blob/master/zenodo/modules/records/data/objecttypes.json
INVENIORDM_TO_CM_TRANSLATIONS = {
    "book": "Book",
    "section": "BookChapter",
    "conferencepaper": "ProceedingsArticle",
    "article": "JournalArticle",
    "patent": "Patent",
    "publication": "JournalArticle",
    "report": "Report",
    "softwaredocumentation": "Software",
    "thesis": "Dissertation",
    "technicalnote": "Report",
    "workingpaper": "Report",
    "datamanagementplan": "OutputManagementPlan",
    "annotationcollection": "Collection",
    "taxonomictreatment": "Collection",
    "peerreview": "PeerReview",
    "poster": "Presentation",
    "presentation": "Presentation",
    "dataset": "Dataset",
    "figure": "Image",
    "plot": "Image",
    "drawing": "Image",
    "photo": "Image",
    "image": "Image",
    "video": "Audiovisual",
    "software": "Software",
    "lesson": "InteractiveResource",
    "physicalobject": "PhysicalObject",
    "workflow": "Workflow",
    "other": "Other",
}

CM_TO_DC_TRANSLATIONS = {
    "Article": "Preprint",
    "Audiovisual": "Audiovisual",
    "Book": "Book",
    "BookChapter": "BookChapter",
    "Collection": "Collection",
    "Dataset": "Dataset",
    "Document": "Text",
    "Entry": "Text",
    "Event": "Event",
    "Figure": "Image",
    "Image": "Image",
    "Instrument": "Instrument",
    "JournalArticle": "JournalArticle",
    "LegalDocument": "Text",
    "Manuscript": "Text",
    "Map": "Image",
    "Patent": "Text",
    "Performance": "Audiovisual",
    "PersonalCommunication": "Text",
    "Post": "Text",
    "ProceedingsArticle": "ConferencePaper",
    "Proceedings": "ConferenceProceeding",
    "Report": "Report",
    "Review": "PeerReview",
    "Software": "Software",
    "Sound": "Sound",
    "Standard": "Standard",
    "StudyRegistration": "StudyRegistration",
    "WebPage": "Text",
}

RIS_TO_CM_TRANSLATIONS = {
    "ABST": "Text",
    "ADVS": "Text",
    "AGGR": "Text",
    "ANCIENT": "Text",
    "ART": "Text",
    "BILL": "Text",
    "BLOG": "Text",
    "BOOK": "Book",
    "CASE": "Text",
    "CHAP": "BookChapter",
    "CHART": "Text",
    "CLSWK": "Text",
    "CTLG": "Collection",
    "COMP": "Software",
    "DATA": "Dataset",
    "DBASE": "Database",
    "DICT": "Dictionary",
    "EBOOK": "Book",
    "ECHAP": "BookChapter",
    "EDBOOK": "Book",
    "EJOUR": "JournalArticle",
    "ELEC": "Text",
    "ENCYC": "Encyclopedia",
    "EQUA": "Equation",
    "FIGURE": "Image",
    "GEN": "CreativeWork",
    "GOVDOC": "GovernmentDocument",
    "GRANT": "Grant",
    "HEAR": "Hearing",
    "ICOMM": "Text",
    "INPR": "Text",
    "JFULL": "JournalArticle",
    "JOUR": "JournalArticle",
    "LEGAL": "LegalRuleOrRegulation",
    "MANSCPT": "Text",
    "MAP": "Map",
    "MGZN": "MagazineArticle",
    "MPCT": "Audiovisual",
    "MULTI": "Audiovisual",
    "MUSIC": "MusicScore",
    "NEWS": "NewspaperArticle",
    "PAMP": "Pamphlet",
    "PAT": "Patent",
    "PCOMM": "PersonalCommunication",
    "RPRT": "Report",
    "SER": "SerialPublication",
    "SLIDE": "Slide",
    "SOUND": "SoundRecording",
    "STAND": "Standard",
    "THES": "Dissertation",
    "UNBILL": "UnenactedBill",
    "UNPB": "UnpublishedWork",
    "VIDEO": "Audiovisual",
    "WEB": "WebPage",
}

CM_TO_RIS_TRANSLATIONS = {
    "Article": "JOUR",
    "Audiovisual": "VIDEO",
    "Book": "BOOK",
    "BookChapter": "CHAP",
    "Collection": "CTLG",
    "Dataset": "DATA",
    "Dissertation": "THES",
    "Document": "GEN",
    "Entry": "DICT",
    "Event": "GEN",
    "Figure": "FIGURE",
    "Image": "FIGURE",
    "JournalArticle": "JOUR",
    "LegalDocument": "GEN",
    "Manuscript": "GEN",
    "Map": "MAP",
    "Patent": "PAT",
    "Performance": "GEN",
    "PersonalCommunication": "PCOMM",
    "Post": "GEN",
    "ProceedingsArticle": "CPAPER",
    "Proceedings": "CONF",
    "Report": "RPRT",
    "Review": "GEN",
    "Software": "COMP",
    "Sound": "SOUND",
    "Standard": "STAND",
    "WebPage": "WEB",
}

SO_TO_CM_TRANSLATIONS = {
    "Article": "Article",
    "BlogPosting": "Article",
    "Book": "Book",
    "BookChapter": "BookChapter",
    "CreativeWork": "Other",
    "Dataset": "Dataset",
    "Dissertation": "Dissertation",
    "Instrument": "Instrument",
    "NewsArticle": "Article",
    "Legislation": "LegalDocument",
    "ScholarlyArticle": "JournalArticle",
    "SoftwareSourceCode": "Software",
}

CM_TO_SO_TRANSLATIONS = {
    "Article": "Article",
    "Audiovisual": "CreativeWork",
    "Book": "Book",
    "BookChapter": "BookChapter",
    "Collection": "CreativeWork",
    "Dataset": "Dataset",
    "Dissertation": "Dissertation",
    "Document": "CreativeWork",
    "Entry": "CreativeWork",
    "Event": "CreativeWork",
    "Figure": "CreativeWork",
    "Image": "CreativeWork",
    "Instrument": "Instrument",
    "JournalArticle": "ScholarlyArticle",
    "LegalDocument": "Legislation",
    "Software": "SoftwareSourceCode",
    "Presentation": "PresentationDigitalDocument",
}

SO_TO_DC_RELATION_TYPES = {
    "citation": "References",
    "isBasedOn": "IsSupplementedBy",
    "sameAs": "IsIdenticalTo",
    "isPartOf": "IsPartOf",
    "hasPart": "HasPart",
    "isPredecessor": "IsPreviousVersionOf",
    "isSuccessor": "IsNewVersionOf",
}

SO_TO_DC_REVERSE_RELATION_TYPES = {
    "citation": "IsReferencedBy",
    "isBasedOn": "IsSupplementTo",
    "sameAs": "IsIdenticalTo",
    "isPartOf": "HasPart",
    "hasPart": "IsPartOf",
    "isPredecessor": "IsNewVersionOf",
    "isSuccessor": "IsPreviousVersionOf",
}

CROSSREF_CONTAINER_TYPES = {
    "book-chapter": "book",
    "dataset": "database",
    "journal-article": "journal",
    "journal-issue": "journal",
    "monograph": "book-series",
    "proceedings-article": "proceedings",
}

COMMONMETA_CONTAINER_TYPES = {
    "Article": "Periodical",
    "BookChapter": "Book",
    "Dataset": "Repository",
    "JournalArticle": "Journal",
    "JournalIssue": "Journal",
    "Monograph": "Book",
    "ProceedingsArticle": "Proceedings",
}

CR_TO_CM_CONTAINER_TRANSLATIONS = {
    "book": "Book",
    "book-series": "BookSeries",
    "database": "DataRepository",
    "journal": "Journal",
    "proceedings": "Proceedings",
}

DC_TO_CM_CONTAINER_TRANSLATIONS = {
    "Book": "Book",
    "BookSeries": "BookSeries",
    "DataRepository": "DataRepository",
    "Journal": "Journal",
    "Periodical": "Periodical",
    "Proceedings": "ProceedingsSeries",
    "Repository": "Repository",
    "Series": "Series",
}

DATACITE_CONTRIBUTOR_TYPES = {
    "ContactPerson": "ContactPerson",
    "DataCollector": "DataCollector",
    "DataCurator": "DataCuration",
    "DataManager": "DataManager",
    "Distributor": "Distributor",
    "Editor": "Editor",
    "HostingInstitution": "HostingInstitution",
    "Other": "Other",
    "Producer": "Producer",
    "ProjectLeader": "ProjectLeader",
    "ProjectManager": "ProjectManager",
    "ProjectMember": "ProjectMember",
    "RegistrationAgency": "RegistrationAgency",
    "RegistrationAuthority": "RegistrationAuthority",
    "RelatedPerson": "RelatedPerson",
    "ResearchGroup": "ResearchGroup",
    "RightsHolder": "RightsHolder",
    "Researcher": "Researcher",
    "Sponsor": "Sponsor",
    "Supervisor": "Supervision",
    "WorkPackageLeader": "WorkPackageLeader",
}

# from commonmeta schema
COMMONMETA_RELATION_TYPES = [
    "IsNewVersionOf",
    "IsPreviousVersionOf",
    "IsVersionOf",
    "HasVersion",
    "IsPartOf",
    "HasPart",
    "IsVariantFormOf",
    "IsOriginalFormOf",
    "IsIdenticalTo",
    "IsTranslationOf",
    "IsReviewedBy",
    "Reviews",
    "IsPreprintOf",
    "HasPreprint",
    "IsSupplementTo",
]

# from commonmeta schema
COMMONMETA_CONTRIBUTOR_ROLES = [
    "Author",
    "Editor",
    "Chair",
    "Reviewer",
    "ReviewAssistant",
    "StatsReviewer",
    "ReviewerExternal",
    "Reader",
    "Translator",
    "ContactPerson",
    "DataCollector",
    "DataManager",
    "Distributor",
    "HostingInstitution",
    "Producer",
    "ProjectLeader",
    "ProjectManager",
    "ProjectMember",
    "RegistrationAgency",
    "RegistrationAuthority",
    "RelatedPerson",
    "ResearchGroup",
    "RightsHolder",
    "Researcher",
    "Sponsor",
    "WorkPackageLeader",
    "Conceptualization",
    "DataCuration",
    "FormalAnalysis",
    "FundingAcquisition",
    "Investigation",
    "Methodology",
    "ProjectAdministration",
    "Resources",
    "Software",
    "Supervision",
    "Validation",
    "Visualization",
    "WritingOriginalDraft",
    "WritingReviewEditing",
    "Maintainer",
    "Other",
]



"""crossref_xml reader for commonmeta-py"""
from typing import Optional
from collections import defaultdict
import httpx
from pydash import py_

from ..utils import (
    doi_from_url,
    dict_to_spdx,
    from_crossref_xml,
    normalize_cc_url,
    normalize_issn,
    normalize_url,
)
from ..base_utils import (
    compact,
    wrap,
    presence,
    sanitize,
    parse_attributes,
    parse_xml,
)
from ..author_utils import get_authors
from ..date_utils import get_date_from_crossref_parts, get_iso8601_date
from ..doi_utils import get_doi_ra, crossref_xml_api_url, normalize_doi
from ..constants import (
    Commonmeta,
    CR_TO_CM_TRANSLATIONS,
    CROSSREF_CONTAINER_TYPES,
    CR_TO_CM_CONTAINER_TRANSLATIONS,
)


def get_crossref_xml(pid: str, **kwargs) -> dict:
    """Get crossref_xml metadata from a DOI"""
    doi = doi_from_url(pid)
    if doi is None:
        return {"state": "not_found"}
    url = crossref_xml_api_url(doi)
    response = httpx.get(
        url, headers={"Accept": "text/xml;charset=utf-8"}, timeout=10, **kwargs
    )
    if response.status_code != 200:
        return {"state": "not_found"}

    return parse_xml(response.text, dialect="crossref")


def read_crossref_xml(data: dict, **kwargs) -> Commonmeta:
    """read_crossref_xml"""
    if data is None:
        return {"state": "not_found"}
    meta = py_.get(
        data, "crossref_result.query_result.body.query.doi_record.crossref", {}
    )

    # query contains information from outside metadata schema, e.g. publisher name
    query = py_.get(data, "crossref_result.query_result.body.query", {})

    # read_options = ActiveSupport::HashWithIndifferentAccess.
    # new(options.except(:doi, :id, :url,
    # :sandbox, :validate, :ra))
    read_options = kwargs or {}

    member_id = next(
        (
            i
            for i in wrap(query.get("crm-item", None))
            if i.get("name", None) == "member-id"
        ),
        {},
    ).get("#text", None)
    publisher_id = (
        "https://api.crossref.org/members/" + member_id if member_id else None
    )
    publisher = compact(
        {
            "id": publisher_id,
            "name": next(
                (
                    i
                    for i in wrap(query.get("crm-item", None))
                    if i.get("name", None) == "publisher-name"
                ),
                {},
            ).get("#text", None),
        }
    )

    # fetch metadata depending of Crossref type
    if py_.get(meta, "journal.journal_article", None):
        bibmeta = py_.get(meta, "journal.journal_article", {})
        resource_type = "journal-article"
    elif py_.get(meta, "journal.journal_issue", None):
        bibmeta = py_.get(meta, "journal.journal_issue", {})
        resource_type = "journal-issue"
    elif py_.get(meta, "journal", None):
        bibmeta = py_.get(meta, "journal", {})
        resource_type = "journal"
    elif py_.get(meta, "posted_content", None):
        bibmeta = meta.get("posted_content", {})
        if publisher.get("name", None) is None:
            publisher = {"name": py_.get(bibmeta, "institution.institution_name", None)}
        resource_type = "posted-content"
    elif py_.get(meta, "book.content_item"):
        bibmeta = py_.get(meta, "book.content_item")
        resource_type = "book-chapter"
    elif py_.get(meta, "book.book_series_metadata"):
        bibmeta = py_.get(meta, "book.book_series_metadata")
        resource_type = "book-series"
    elif py_.get(meta, "book.book_set_metadata"):
        bibmeta = py_.get(meta, "book.book_set_metadata")
        resource_type = "book-set"
    elif py_.get(meta, "book.book_metadata"):
        bibmeta = py_.get(meta, "book.book_metadata")
        resource_type = "book"
    elif py_.get(meta, "conference", None):
        bibmeta = py_.get(meta, "conference.conference_paper", {})
        resource_type = "proceedings-article"
    elif py_.get(meta, "sa_component", None):
        bibmeta = py_.get(meta, "sa_component.component_list.component", {})
        resource_type = "component"
    elif py_.get(meta, "database", None):
        bibmeta = py_.get(meta, "database.dataset", {})
        resource_type = "dataset"
    elif py_.get(meta, "report_paper", None):
        bibmeta = py_.get(meta, "report_paper.report_paper_metadata", {})
        resource_type = "report"
    elif py_.get(meta, "peer_review", None):
        bibmeta = py_.get(meta, "peer_review", {})
        resource_type = "peer-review"
    elif py_.get(meta, "dissertation", None):
        bibmeta = py_.get(meta, "dissertation", {})
        resource_type = "dissertation"
    else:
        bibmeta = {}
        resource_type = ""

    _id = normalize_doi(
        kwargs.get("doi", None)
        or kwargs.get("id", None)
        or py_.get(bibmeta, "doi_data.doi")
    )
    _type = CR_TO_CM_TRANSLATIONS.get(resource_type, "Other")
    url = parse_attributes(py_.get(bibmeta, "doi_data.resource"))
    url = normalize_url(url)
    titles = crossref_titles(bibmeta)
    contributors = crossref_people(bibmeta)

    date: dict = defaultdict(list)
    date["created"] = next(
        (
            i
            for i in wrap(query.get("crm-item", None))
            if i.get("name", None) == "created"
        ),
        {},
    ).get("#text", None)
    date["published"] = (
        get_date_from_crossref_parts(bibmeta.get("publication_date", {}))
        or get_date_from_crossref_parts(bibmeta.get("review_date", {}))
        or date["created"]
    )
    date["updated"] = next(
        (
            i
            for i in wrap(query.get("crm-item", None))
            if i.get("name", None) == "last-update"
        ),
        {},
    ).get("#text", None)

    # TODO: fix timestamp. Until then, remove time as this is not always stable with Crossref (different server timezones)
    date = {k: get_iso8601_date(v) for k, v in date.items()}

    descriptions = crossref_description(bibmeta)
    funding = (
        py_.get(bibmeta, "program")
        or py_.get(bibmeta, "program.0.assertion")
        or py_.get(bibmeta, "crossmark.custom_metadata.program.assertion")
        or py_.get(bibmeta, "crossmark.custom_metadata.program.0.assertion")
    )
    funding_references = crossref_funding(wrap(funding))

    license_ = (
        py_.get(bibmeta, "program.license_ref")
        or py_.get(bibmeta, "crossmark.custom_metadata.program.license_ref")
        or py_.get(bibmeta, "crossmark.custom_metadata.program.1.license_ref")
    )
    license_ = crossref_license(wrap(license_))

    # By using book_metadata, we can account for where resource_type is `BookChapter` and not assume its a whole book
    # if book_metadata:
    #     #   identifiers = crossref_alternate_identifiers(book_metadata)
    #     container = compact(
    #         {
    #             "type": "Book",
    #             "title": py_.get(book_metadata, "titles.title"),
    #             "firstPage": py_.get(bibmeta, "pages.first_page"),
    #             "lastPage": py_.get(bibmeta, "pages.last_page"),
    #             #'identifiers' => identifiers
    #         }
    #     )

    # elif book_series_metadata.get("series_metadata", None):
    #     issn = normalize_issn(
    #         py_.get(book_series_metadata, "series_metadata.issn.0.#text")
    #     )
    #     container = compact(
    #         {
    #             "type": "Book Series",
    #             "identifier": issn,
    #             "identifierType": "ISSN" if issn else None,
    #             "title": py_.get(book_series_metadata, "series_metadata.titles.title"),
    #             "volume": bibmeta.get("volume", None),
    #         }
    #     )
    # else:
    #     container = None
    container = crossref_container(meta, resource_type=resource_type)
    references = [
        crossref_reference(i) for i in wrap(py_.get(bibmeta, "citation_list.citation"))
    ]
    language = py_.get(meta, "journal.journal_metadata.language")

    files = presence(meta.get("contentUrl", None))
    provider = (
        bibmeta.get("reg-agency").capitalize()
        if bibmeta.get("reg-agency", None)
        else None
    )
    if provider is None:
        provider = get_doi_ra(_id)
    state = "findable" if meta or read_options else "not_found"

    return {
        # required properties
        "id": _id,
        "type": _type,
        "url": url,
        "contributors": presence(contributors),
        "titles": presence(titles),
        "publisher": publisher,
        "date": compact(date),
        # recommended and optional properties
        "subjects": presence(None),
        "language": language,
        "alternate_identifiers": None,
        "sizes": None,
        "formats": None,
        "version": None,
        "license": presence(license_),
        "descriptions": presence(descriptions),
        "geo_locations": None,
        "funding_references": presence(funding_references),
        "references": references,
        # other properties
        "date_created": None,
        "date_registered": None,
        "date_published": None,
        "date_updated": None,
        "content_url": presence(files),
        "container": presence(container),
        "provider": provider,
        "state": state,
        "schema_version": None,
    } | read_options


def crossref_titles(bibmeta):
    """Title information from Crossref metadata."""
    title = parse_attributes(py_.get(bibmeta, "titles.0.title"))
    subtitle = parse_attributes(py_.get(bibmeta, "titles.0.subtitle"))
    original_language_title = parse_attributes(
        py_.get(bibmeta, "titles.0.original_language_title")
    )
    language = parse_attributes(
        py_.get(bibmeta, "titles.0.original_language_title"), content="language"
    )
    if title is None and original_language_title is None:
        return None
    if title and original_language_title is None and subtitle is None:
        return [{"title": sanitize(title)}]
    if original_language_title:
        return [
            compact(
                {
                    "title": sanitize(original_language_title),
                    "lang": language,
                }
            )
        ]
    if subtitle:
        return [
            compact({"title": sanitize(title)}),
            {
                "title": sanitize(subtitle),
                "titleType": "Subtitle",
            },
        ]


def crossref_description(bibmeta):
    """Description information from Crossref metadata."""

    def format_abstract(element):
        """Format abstract"""
        if isinstance(element.get("p", None), list):
            element["p"] = element["p"][0]
        if isinstance(element.get("p", None), dict):
            element["p"] = element["p"]["#text"]
        description_type = (
            "Abstract" if element.get("abstract-type", None) == "abstract" else "Other"
        )
        return compact(
            {
                "descriptionType": description_type,
                "description": sanitize(
                    parse_attributes(element, content="p", first=True)
                ),
            }
        )

    return [format_abstract(i) for i in wrap(bibmeta.get("abstract", None))]


def crossref_people(bibmeta):
    """Person information from Crossref metadata."""

    person = py_.get(bibmeta, "contributors.person_name") or bibmeta.get(
        "person_name", None
    )
    organization = wrap(py_.get(bibmeta, "contributors.organization"))

    return get_authors(from_crossref_xml(wrap(person) + wrap(organization)))

    #     (Array.wrap(person) + Array.wrap(organization)).select do |a|
    #       a['contributor_role'] == contributor_role
    #     end.map do |a|
    #       name_identifiers = if normalize_orcid(parse_attributes(a['ORCID'])).present?
    #                            [{
    #                              'nameIdentifier' => normalize_orcid(parse_attributes(a['ORCID'])), 'nameIdentifierScheme' => 'ORCID', 'schemeUri' => 'https://orcid.org'
    #                            }]
    #                          end
    #       if a['surname'].present? || a['given_name'].present? || name_identifiers.present?
    #         given_name = parse_attributes(a['given_name'])
    #         family_name = parse_attributes(a['surname'])
    #         affiliation = Array.wrap(a['affiliation']).map do |a|
    #           if a.is_a?(Hash)
    #             a
    #           elsif a.is_a?(Hash) && a.key?('#text') && a[#text'].strip.blank?
    #             nil
    #           elsif a.is_a?(Hash) && a.key?('_#text_')
    #             { 'name' => a['#text'] }
    #           elsif a.strip.blank?
    #             nil
    #           elsif a.is_a?(String)
    #             { 'name' => a }
    #           end
    #         end.compact

    #         { 'nameType' => 'Personal',
    #           'nameIdentifiers' => name_identifiers,
    #           'name' => [family_name, given_name].compact.join(', '),
    #           'givenName' => given_name,
    #           'familyName' => family_name,
    #           'affiliation' => affiliation.presence,
    #           'contributorType' => contributor_role == 'editor' ? 'Editor' : nil }.compact
    #       else
    #         { 'nameType' => 'Organizational',
    #           'name' => a['name'] || a['#text'] }


def crossref_reference(reference: Optional[dict]) -> Optional[dict]:
    """Get reference from Crossref reference"""
    if reference is None or not isinstance(reference, dict):
        return None
    doi = parse_attributes(reference.get("doi", None))
    unstructured = reference.get("unstructured_citation", None)
    if isinstance(unstructured, dict):
        url = unstructured.get("u", None)
        text = unstructured.get("font", None) or unstructured.get("#text", None)
    else:
        url = reference.get("url", None)
        text = reference.get("unstructured_citation", None)
    metadata = {
        "key": reference.get("key", None),
        "doi": normalize_doi(doi) if doi else None,
        "url": normalize_url(url) if url else None,
        "contributor": reference.get("author", None),
        "title": reference.get("article_title", None),
        "publisher": reference.get("publisher", None),
        "publicationYear": reference.get("cYear", None),
        "volume": reference.get("volume", None),
        "issue": reference.get("issue", None),
        "firstPage": reference.get("first_page", None),
        "lastPage": reference.get("last_page", None),
        "containerTitle": reference.get("journal_title", None),
        "edition": None,
        "unstructured": sanitize(text) if text and doi is None else None,
    }
    return compact(metadata)


def crossref_container(meta: dict, resource_type: str = "JournalArticle") -> dict:
    """Get container from Crossref"""
    container_type = CROSSREF_CONTAINER_TYPES.get(resource_type, None)
    issn = next(
        (
            i
            for i in wrap(
                py_.get(meta, f"{container_type}.{container_type}_metadata.issn")
            )
            + wrap(
                py_.get(
                    meta,
                    f"{container_type}.{container_type}_series_metadata.series_metadata.issn",
                )
            )
            if i.get("media_type", None) == "electronic"
        ),
        {},
    ) or next(
        (
            i
            for i in wrap(
                py_.get(meta, f"{container_type}.{container_type}_metadata.issn")
            )
            + wrap(
                py_.get(
                    meta,
                    f"{container_type}.{container_type}_series_metadata.series_metadata.issn",
                )
            )
            if i.get("media_type", None) == "print"
        ),
        {},
    )
    issn = normalize_issn(issn) if issn else None
    isbn = py_.get(meta, f"conference.{container_type}_metadata.isbn.#text")
    container_title = (
        py_.get(meta, f"{container_type}.{container_type}_metadata.full_title")
        or py_.get(meta, f"{container_type}.{container_type}_metadata.titles.0.title")
        or py_.get(meta, f"conference.{container_type}_metadata.{container_type}_title")
        or py_.get(
            meta,
            f"{container_type}.{container_type}_series_metadata.series_metadata.titles.0.title",
        )
    )
    volume = py_.get(
        meta,
        f"{container_type}.{container_type}_issue.{container_type}_volume.volume",
    )
    issue = py_.get(meta, f"{container_type}.{container_type}_issue.issue")
    return compact(
        {
            "type": CR_TO_CM_CONTAINER_TRANSLATIONS.get(container_type, None),
            "identifier": issn or isbn,
            "identifierType": "ISSN" if issn else "ISBN" if isbn else None,
            "title": container_title,
            "volume": volume,
            "issue": issue,
            "firstPage": py_.get(
                meta, f"{container_type}.{container_type}_article.pages.first_page"
            )
            or py_.get(meta, f"{container_type}.content_item.pages.first_page")
            or py_.get(meta, "conference.conference_paper.pages.first_page"),
            "lastPage": py_.get(
                meta, f"{container_type}.{container_type}_article.pages.last_page"
            )
            or py_.get(meta, f"{container_type}.content_item.pages.last_page")
            or py_.get(meta, "conference.conference_paper.pages.last_page"),
            "location": py_.get(meta, "conference.event_metadata.conference_location"),
            "series": py_.get(meta, "conference.event_metadata.conference_acronym"),
        }
    )


def crossref_funding(funding: list) -> list:
    """Get assertions from Crossref"""
    return []


def crossref_license(licenses: list) -> dict:
    """Get license from Crossref"""

    def map_element(element):
        """Format element"""
        url = parse_attributes(element)
        url = normalize_cc_url(url)
        return dict_to_spdx({"url": url})

    # return only the first license found
    return next((map_element(i) for i in licenses), None)



"""datacite reader for Commonmeta"""
from collections import defaultdict
from typing import Optional
import httpx
from pydash import py_

from ..utils import (
    normalize_url,
    normalize_doi,
    normalize_cc_url,
    dict_to_spdx,
    format_name_identifier,
)
from ..base_utils import compact, wrap, presence
from ..author_utils import get_authors
from ..date_utils import normalize_date_dict
from ..doi_utils import (
    doi_as_url,
    doi_from_url,
    datacite_api_url,
    datacite_api_sample_url,
)
from ..constants import (
    DC_TO_CM_TRANSLATIONS,
    DC_TO_CM_CONTAINER_TRANSLATIONS,
    Commonmeta,
)


def get_datacite(pid: str, **kwargs) -> dict:
    """get_datacite"""
    doi = doi_from_url(pid)
    if doi is None:
        return {"state": "not_found"}
    url = datacite_api_url(doi)
    try:
        response = httpx.get(url, timeout=10, **kwargs)
        if response.status_code != 200:
            return {"state": "not_found"}
        return py_.get(response.json(), "data.attributes", {})
    except httpx.ReadTimeout:
        return {"state": "timeout"}


def read_datacite(data: dict, **kwargs) -> Commonmeta:
    """read_datacite"""
    meta = data
    if data is None:
        return {"state": "not_found"}

    read_options = kwargs or {}

    _id = doi_as_url(meta.get("doi", None))
    resource__typegeneral = py_.get(meta, "types.resourceTypeGeneral")
    resource_type = py_.get(meta, "types.resourceType")
    _type = DC_TO_CM_TRANSLATIONS.get(resource__typegeneral, "Other")
    additional_type = DC_TO_CM_TRANSLATIONS.get(resource_type, None)
    # if resource_type is one of the new resource__typegeneral types introduced in schema 4.3, use it
    if additional_type:
        _type = additional_type
        additional_type = None
    else:
        additional_type = resource_type
    titles = get_titles(wrap(meta.get("titles", None)))

    contributors = get_authors(wrap(meta.get("creators", None)))
    contrib = get_authors(wrap(meta.get("contributors", None)))
    if contrib:
        contributors = contributors + contrib

    publisher = meta.get("publisher", None)
    if isinstance(publisher, str):
        publisher = {"name": publisher}
    elif isinstance(publisher, dict):
        publisher = get_publisher(publisher)
    date = get_dates(wrap(meta.get("dates", None)), meta.get("publicationYear", None))
    container = get_container(meta.get("container", None))
    license_ = meta.get("rightsList", [])
    if len(license_) > 0:
        license_ = normalize_cc_url(license_[0].get("rightsUri", None))
        license_ = dict_to_spdx({"url": license_}) if license_ else None

    files = [get_file(i) for i in wrap(meta.get("content_url"))]
    references = get_references(
        wrap(meta.get("relatedItems", None) or meta.get("relatedIdentifiers", None))
    )
    descriptions = get_descriptions(wrap(meta.get("descriptions", None)))
    geo_locations = get_geolocation(wrap(meta.get("geoLocations", None)))
    formats = py_.uniq(meta.get("formats", None))

    return {
        # required properties
        "id": _id,
        "type": _type,
        "doi": doi_from_url(_id),
        "url": normalize_url(meta.get("url", None)),
        "contributors": presence(contributors),
        "titles": titles,
        "publisher": publisher,
        "date": compact(date),
        # recommended and optional properties
        "additional_type": additional_type,
        "subjects": presence(meta.get("subjects", None)),
        "language": meta.get("language", None),
        "alternate_identifiers": presence(meta.get("alternateIdentifiers", None)),
        "sizes": presence(meta.get("sizes", None)),
        "formats": presence(formats),
        "version": meta.get("version", None),
        "license": presence(license_),
        "descriptions": descriptions,
        "geo_locations": presence(geo_locations),
        "funding_references": presence(meta.get("fundingReferences", None)),
        "references": presence(references),
        # other properties
        "files": presence(files),
        "container": presence(container),
        "provider": "DataCite",
        "state": "findable",
        "schema_version": meta.get("schemaVersion", None),
    } | read_options


def get_references(references: list) -> list:
    """get_references"""

    def is_reference(reference):
        """is_reference"""
        return reference.get("relationType", None) in ["Cites", "References"]

    def map_reference(reference, index):
        """map_reference"""
        identifier = reference.get("relatedIdentifier", None)
        identifier_type = reference.get("relatedIdentifierType", None)
        return compact(
            {
                "key": f"ref{index + 1}",
                "doi": normalize_doi(identifier) if identifier_type == "DOI" else None,
                "url": normalize_url(identifier) if identifier_type == "URL" else None,
            }
        )

    return [
        map_reference(i, index) for index, i in enumerate(references) if is_reference(i)
    ]


def get_file(file: str) -> dict:
    """get_file"""
    return compact({"url": file})


def get_dates(dates: list, publication_year) -> dict:
    """convert date list to dict, rename and/or remove some keys"""
    date: dict = defaultdict(list)
    for sub in dates:
        date[sub.get("dateType", None)] = sub.get("date", None)
    if date.get("Issued", None) is None and publication_year is not None:
        date["Issued"] = str(publication_year)
    return normalize_date_dict(date)


def get_descriptions(descriptions: list) -> list:
    """get_descriptions"""

    def map_description(description):
        """map_description"""
        return compact(
            {
                "description": description.get("description", None),
                "type": description.get("descriptionType")
                if description.get("descriptionType", None)
                in ["Abstract", "Methods", "TechnicalInfo", "Other"]
                else "Other",
                "language": description.get("lang", None),
            }
        )

    return [
        map_description(i)
        for i in descriptions
        if i.get("description", None) is not None
    ]


def get_titles(titles: list) -> list:
    """get_titles"""

    def map_title(title):
        """map_title"""
        return compact(
            {
                "title": title.get("title", None),
                "type": title.get("titleType")
                if title.get("titleType", None)
                in ["AlternativeTitle", "Subtitle", "TranslatedTitle"]
                else None,
                "language": title.get("lang", None),
            }
        )

    return [map_title(i) for i in titles if i.get("title", None) is not None]


def get_publisher(publisher: dict) -> dict:
    """get_publisher"""
    return compact(
        {"id": format_name_identifier(publisher), "name": publisher.get("name", None)}
    )


def get_geolocation(geolocations: list) -> list:
    """get_geolocation"""

    def geo_location_point(point: dict):
        """geo_location_point, convert lat and long to int"""
        return {
            "pointLatitude": float(point.get("pointLatitude"))
            if point.get("pointLatitude", None)
            else None,
            "pointLongitude": float(point.get("pointLongitude"))
            if point.get("pointLongitude", None)
            else None,
        }

    def geo_location_box(box: dict):
        """geo_location_box, convert lat and long to int"""
        return {
            "eastBoundLongitude": float(box.get("eastBoundLongitude"))
            if box.get("eastBoundLongitude", None)
            else None,
            "northBoundLatitude": float(box.get("northBoundLatitude"))
            if box.get("northBoundLatitude", None)
            else None,
            "southBoundLatitude": float(box.get("southBoundLatitude"))
            if box.get("southBoundLatitude", None)
            else None,
            "westBoundLongitude": float(box.get("westBoundLongitude"))
            if box.get("westBoundLongitude", None)
            else None,
        }

    return [
        compact(
            {
                "geoLocationPoint": geo_location_point(location.get("geoLocationPoint"))
                if location.get("geoLocationPoint", None)
                else None,
                "geoLocationBox": geo_location_box(location.get("geoLocationBox"))
                if location.get("geoLocationBox", None)
                else None,
                "geoLocationPlace": location.get("geoLocationPlace", None),
            }
        )
        for location in geolocations
    ]


def get_container(container: Optional[dict]) -> dict or None:
    """get_container"""
    if container is None:
        return None
    _type = (
        DC_TO_CM_CONTAINER_TRANSLATIONS.get(container.get("type"), None)
        if container.get("type", None)
        else None
    )

    return compact(
        {
            "id": container.get("identifier", None),
            "type": _type,
            "title": container.get("title", None),
        }
    )


def get_random_datacite_id(number: int = 1) -> list:
    """Get random DOI from DataCite"""
    number = 20 if number > 20 else number
    url = datacite_api_sample_url(number)
    try:
        response = httpx.get(url, timeout=60)
        if response.status_code != 200:
            return []

        items = py_.get(response.json(), "data")
        return [i.get("id") for i in items]
    except httpx.ReadTimeout:
        return []



"""Date utils for commonmeta-py"""
import datetime
from datetime import datetime as dt
from typing import Optional, Union
import dateparser
import pydash as py_

from .base_utils import compact

MONTH_NAMES = {
    "01": "jan",
    "02": "feb",
    "03": "mar",
    "04": "apr",
    "05": "may",
    "06": "jun",
    "07": "jul",
    "08": "aug",
    "09": "sep",
    "10": "oct",
    "11": "nov",
    "12": "dec",
}

MONTH_SHORT_NAMES = [
    "jan",
    "feb",
    "mar",
    "apr",
    "may",
    "jun",
    "jul",
    "aug",
    "sep",
    "oct",
    "nov",
    "dec",
]

ISO8601_DATE_FORMAT = "%Y-%m-%d"


def get_iso8601_date(date: Union[datetime.datetime, datetime.date, str, int]) -> str:
    """Get ISO 8601 date without time"""
    if date is None:
        return ""
    if isinstance(date, (datetime.datetime, datetime.date)):
        return date.strftime(ISO8601_DATE_FORMAT)
    if isinstance(date, str):
        length = len(date)
        if length == 7:
            return dateparser.parse(date).strftime("%Y-%m")
        if length == 4:
            return dateparser.parse(date).strftime("%Y")
        else:
            return dateparser.parse(date).strftime(ISO8601_DATE_FORMAT)
    if isinstance(date, int):
        return datetime.datetime.fromtimestamp(date).strftime(ISO8601_DATE_FORMAT)
    return ""


def get_date_parts(iso8601_time: Optional[str]) -> dict:
    """Get date parts"""
    if iso8601_time is None:
        return {"date-parts": [[]]}

    # add 0s to the end of the date if it is incomplete
    if len(iso8601_time) < 10:
        iso8601_time = iso8601_time.ljust(10, "0")

    year = int(iso8601_time[0:4])
    month = int(iso8601_time[5:7])
    day = int(iso8601_time[8:10])

    date_parts = py_.reject([year, month, day], lambda x: x == 0)
    return {"date-parts": [date_parts]}


def get_date_from_unix_timestamp(timestamp: Optional[int]) -> Optional[str]:
    """Get date from unix timestamp"""
    if timestamp is None:
        return None
    return datetime.datetime.fromtimestamp(timestamp).replace(microsecond=0).isoformat()


def get_date_from_date_parts(date_as_parts: Optional[dict]) -> Optional[str]:
    """Get date from date parts"""
    if date_as_parts is None:
        return None
    date_parts = date_as_parts.get("date-parts", [])
    if len(date_parts) == 0:
        return None
    date_parts = date_parts[0]
    if date_parts[0] is None:
        return None
    year = date_parts[0] if len(date_parts) > 0 else 0
    month = date_parts[1] if len(date_parts) > 1 else 0
    day = date_parts[2] if len(date_parts) > 2 else 0
    return get_date_from_parts(year, month, day)


def get_date_from_crossref_parts(date_parts: dict):
    """Get date from Crossref XML date parts"""
    if isinstance(date_parts, list):
        date_parts = date_parts[0]
    year = date_parts.get("year", None)
    if year is None:
        return None
    month = date_parts.get("month", 0)
    day = date_parts.get("day", 0)
    return get_date_from_parts(year, month, day)


def get_date_from_parts(year=0, month=0, day=0) -> Optional[str]:
    """Get date from parts"""
    arr = [str(year).rjust(4, "0"), str(month).rjust(2, "0"), str(day).rjust(2, "0")]
    arr = [e for i, e in enumerate(arr) if (e not in ["00", "0000"])]
    return None if len(arr) == 0 else "-".join(arr)


def get_month_from_date(
    date: Optional[Union[str, int, datetime.datetime, datetime.date]],
) -> Optional[str]:
    """Get month from date"""
    if date is None:
        return None
    # if date type is not recognized
    if not isinstance(date, (str, int, datetime.datetime, datetime.date)):
        return None
    if isinstance(date, str):
        date = dateparser.parse(date).strftime(ISO8601_DATE_FORMAT)
    if isinstance(date, int):
        date = datetime.datetime.fromtimestamp(date).strftime(ISO8601_DATE_FORMAT)
    if isinstance(date, (datetime.datetime, datetime.date)):
        date = date.strftime(ISO8601_DATE_FORMAT)
    date = date.split("-")
    return MONTH_NAMES.get(date[1], None) if len(date) > 1 else None


def strip_milliseconds(iso8601_time: Optional[str]) -> Optional[str]:
    """strip milliseconds if there is a time, as it interferes with edtc parsing"""
    if iso8601_time is None or len(iso8601_time) == 0:
        return None
    if "T00:00:00" in iso8601_time:
        return iso8601_time.split("T")[0]
    if "." in iso8601_time:
        return iso8601_time.split(".")[0] + "Z"
    if "+00:00" in iso8601_time:
        return iso8601_time.split("+")[0] + "Z"
    return iso8601_time


def get_datetime_from_time(time: str) -> Optional[str]:
    """iso8601 datetime without hyphens and colons, used by Crossref"""
    try:
        return dt.strptime(time, "%Y%m%d%H%M%S").strftime("%Y-%m-%dT%H:%M:%SZ")
    except ValueError:
        return None


def normalize_date_dict(data: dict) -> dict:
    """Normalize date dict

    Supported date types in commonmeta:
    - created
    - submitted
    - accepted
    - published
    - available
    - updated
    - withdrawn
    """
    return compact(
        {
            "created": data.get("Created", None),
            "submitted": data.get("Submitted", None),
            "accepted": data.get("Accepted", None),
            "published": data.get("Issued", None),
            "available": data.get("Available", None),
            "updated": data.get("Updated", None),
            "withdrawn": data.get("Withdrawn", None),
        }
    )



"""RIS reader for commonmeta-py"""
from typing import Optional

from ..utils import compact, normalize_url, wrap
from ..base_utils import presence
from ..author_utils import get_authors
from ..date_utils import get_date_from_parts
from ..doi_utils import normalize_doi, doi_from_url
from ..constants import RIS_TO_CM_TRANSLATIONS, Commonmeta


def read_ris(data: Optional[str], **kwargs) -> Commonmeta:
    """read_ris"""

    meta = ris_meta(data=data)
    read_options = kwargs or {}

    if not isinstance(meta, dict):
        return {"state": "not_found"}

    _id = read_options.get("doi", None) or normalize_doi(meta.get("DO", None))
    _type = RIS_TO_CM_TRANSLATIONS.get(meta.get("TY", None), "Other")
    container_type = "Journal" if _type == "JournalArticle" else None

    def get_author(author):
        """get_author"""
        return {"creatorName": author}

    authors = [get_author(i) for i in wrap(meta.get("AU", None))]
    contributors = get_authors(authors)
    date = {}
    if meta.get("PY", None) is not None:
        date["published"] = get_date_from_parts(*str(meta.get("PY", None)).split("/"))
    if meta.get("Y1", None) is not None:
        date["created"] = get_date_from_parts(*str(meta.get("Y1", None)).split("/"))
    # related_identifiers = if meta.fetch('T2', nil).present? & & meta.fetch('SN', nil).present?
    #                             [{'type' = > 'Periodical',
    #                                'id'= > meta.fetch('SN', nil),
    #                                'relatedIdentifierType'= > 'ISSN',
    #                                'relationType'= > 'IsPartOf',
    #                                'title' = > meta.fetch('T2', nil)}.compact]
    #                           else
    #                             []
    #                           end
    descriptions = None
    if meta.get("AB", None) is not None:
        descriptions = [{"description": meta.get("AB"), "descriptionType": "Abstract"}]
    if meta.get("T2", None) is not None:
        container = compact(
            {
                "type": container_type,
                "title": meta.get("T2", None),
                "volume": meta.get("VL", None),
                "issue": meta.get("IS", None),
                "firstPage": meta.get("SP", None),
                "lastPage": meta.get("EP", None),
            }
        )
    else:
        container = None
    if meta.get("PB", None) is not None:
        publisher = {"name": meta.get("PB")}
    else:
        publisher = None
    subjects = wrap(meta.get("KW", None))
    state = "findable" if meta.get("DO", None) or read_options else "not_found"

    return {
        "id": _id,
        "type": _type,
        "doi": doi_from_url(_id),
        "url": normalize_url(meta.get("UR", None)),
        "titles": [{"title": meta.get("T1", None)}],
        "descriptions": descriptions,
        "contributors": presence(contributors),
        "publisher": presence(publisher),
        "container": container,
        # 'related_identifiers': related_identifiers,
        "date": date,
        "subjects": subjects,
        "language": meta.get("LA", None),
        "state": state,
    } | read_options


def ris_meta(data):
    """ris_meta"""
    meta = {}
    if data is None:
        return meta
    for line in data.split("\n"):
        values = line.split("-", 2)
        key = values[0].strip()
        if len(values) == 1:
            continue
        if meta.get(key, None) is None:
            meta[key] = values[1].strip()
        elif isinstance(meta[key], str):
            meta[key] = [meta[key]]
        elif isinstance(meta[key], list):
            meta[key].append(values[1].strip())

    return meta



"""API Utils module for commonmeta-py"""

from typing import Optional
from datetime import datetime as date
import httpx
from furl import furl
import jwt

from .doi_utils import validate_doi, doi_as_url
from .readers.json_feed_reader import get_json_feed_item_uuid


def generate_ghost_token(key: str) -> str:
    """Generate a short-lived JWT for the Ghost Admin API.
    From https://ghost.org/docs/admin-api/#token-authentication"""

    # Split the key into ID and SECRET
    _id, secret = key.split(":")

    # Prepare header and payload
    iat = int(date.now().timestamp())

    header = {"alg": "HS256", "typ": "JWT", "kid": _id}
    payload = {"iat": iat, "exp": iat + 5 * 60, "aud": "/admin/"}

    # Create and return the token (including decoding secret)
    return jwt.encode(payload, bytes.fromhex(secret), algorithm="HS256", headers=header)


def update_ghost_post_via_api(
    _id: str, api_key: Optional[str] = None, api_url: Optional[str] = None
) -> dict[str, str]:
    """Update Ghost post via API"""
    # get post doi and url from Rogue Scholar API
    # post url is needed to find post via Ghost API
    post = get_json_feed_item_uuid(_id)
    if post.get("error", None):
        return post
    doi = validate_doi(post.get("doi", None))
    doi = doi_as_url(doi)
    url = post.get("url", None)
    if not doi or not url:
        return {"error": "DOI or URL not found"}

    # get post_id and updated_at from ghost api
    token = generate_ghost_token(api_key)
    headers = {
        "Authorization": f"Ghost {token}",
        "Content-Type": "application/json",
        "Accept-Version": "v5",
    }
    f = furl(url)
    slug = f.path.segments[-1]
    ghost_url = f"{api_url}/ghost/api/admin/posts/slug/{slug}/"
    response = httpx.get(ghost_url, headers=headers, timeout=10)
    if response.status_code != 200:
        return {"error": "Error fetching post"}
    ghost_post = response.json().get("posts")[0]
    guid = ghost_post.get("id")
    updated_at = ghost_post.get("updated_at")
    if not guid or not updated_at:
        return {"error": "guid or updated_at not found"}

    # update post canonical_url using doi. This requires sending
    # the updated_at timestamp to avoid conflicts, and must use guid
    # rather than url for put requests
    ghost_url = f"{api_url}/ghost/api/admin/posts/{guid}/"

    json = {"posts": [{"canonical_url": doi, "updated_at": updated_at}]}
    response = httpx.put(
        ghost_url,
        headers=headers,
        json=json,
    )
    if response.status_code != 200:
        return {"error": "Error updating post"}
    return {"message": f"DOI {doi} added", "guid": guid, "updated_at": updated_at}



"""Crossref XML writer for commonmeta-py"""
from typing import Optional
from ..constants import Commonmeta
from ..crossref_utils import generate_crossref_xml, generate_crossref_xml_list


def write_crossref_xml(metadata: Commonmeta) -> Optional[str]:
    """Write Crossref XML"""
    return generate_crossref_xml(metadata)


def write_crossref_xml_list(metalist):
    """Write crossref_xml list"""
    if metalist is None:
        return None

    return generate_crossref_xml_list(metalist)



"""schema_org reader for commonmeta-py"""
from typing import Optional
import orjson as json
from collections import defaultdict
import httpx
from pydash import py_
from bs4 import BeautifulSoup

from ..utils import (
    dict_to_spdx,
    normalize_cc_url,
    from_schema_org,
    from_schema_org_creators,
    normalize_id,
    normalize_ids,
    normalize_url,
    name_to_fos,
)
from ..base_utils import wrap, compact, presence, parse_attributes, sanitize
from ..author_utils import get_authors
from ..date_utils import get_iso8601_date, strip_milliseconds
from ..doi_utils import doi_from_url, get_doi_ra
from ..constants import (
    SO_TO_CM_TRANSLATIONS,
    SO_TO_DC_RELATION_TYPES,
    SO_TO_DC_REVERSE_RELATION_TYPES,
    Commonmeta,
)


def get_schema_org(pid: str, **kwargs) -> dict:
    """get_schema_org"""
    if pid is None:
        return {"state": "not_found"}
    url = pid
    response = httpx.get(url, timeout=10, follow_redirects=True, **kwargs)
    if response.status_code != 200:
        return {"state": "not_found"}

    soup = BeautifulSoup(response.text, "html.parser")
    # workaround for metadata not included with schema.org but in html meta tags
    data = get_html_meta(soup)
    # load schema.org metadata
    json_ld = soup.find("script", type="application/ld+json")
    if json_ld is not None:
        data |= json.loads(json_ld.text)

    # workaround if not all authors are included with schema.org (e.g. in Ghost metadata)
    auth = soup.select("meta[name='citation_author']")

    def format_author(author):
        return {"name": author["content"]}

    authors = [format_author(i) for i in auth]

    if data.get("author", None) is None and data.get("creator", None) is not None:
        data["author"] = data["creator"]
    if len(authors) > len(wrap(data.get("author", None))):
        data["author"] = authors

    return data


def read_schema_org(data: Optional[dict], **kwargs) -> Commonmeta:
    """read_schema_org"""
    if data is None:
        return {"meta": None, "state": "not_found"}
    meta = data

    read_options = kwargs or {}

    _id = meta.get("@id", None)
    if _id is None:
        _id = meta.get("identifier", None)
    _id = normalize_id(_id)
    _type = SO_TO_CM_TRANSLATIONS.get(meta.get("@type", None), "Other")
    additional_type = meta.get("additionalType", None)
    authors = meta.get("author", None) or meta.get("creator", None)
    # Authors should be an object, if it's just a plain string don't try and parse it.
    if not isinstance(authors, str):
        contributors = get_authors(from_schema_org_creators(wrap(authors)))
    else:
        contributors = authors
    contrib = presence(
        get_authors(from_schema_org_creators(wrap(meta.get("editor", None))))
    )
    if contrib:
        contributors = contributors + contrib

    if meta.get("name", None) is not None:
        titles = [{"title": meta.get("name")}]
    elif meta.get("headline", None) is not None:
        titles = [{"title": meta.get("headline")}]
    else:
        titles = None

    date: dict = defaultdict(list)
    date["published"] = strip_milliseconds(meta.get("datePublished", None))
    date["updated"] = strip_milliseconds(meta.get("dateModified", None))

    publisher = meta.get("publisher", None)
    if publisher is not None:
        publisher = py_.omit(
            publisher, ["@type", "logo", "url", "disambiguatingDescription"]
        )

    license_ = meta.get("license", None)
    if license_ is not None:
        license_ = normalize_cc_url(license_)
        license_ = dict_to_spdx({"url": license_}) if license_ else None

    if _type == "Dataset":
        url = parse_attributes(
            from_schema_org(meta.get("includedInDataCatalog", None)),
            content="url",
            first=True,
        )
        container = compact(
            {
                "type": "DataRepository",
                "title": parse_attributes(
                    from_schema_org(meta.get("includedInDataCatalog", None)),
                    content="name",
                    first=True,
                ),
                "identifier": url,
                "identifierType": "URL" if url is not None else None,
                "volume": meta.get("volumeNumber", None),
                "issue": meta.get("issueNumber", None),
                "firstPage": meta.get("pageStart", None),
                "lastPage": meta.get("pageEnd", None),
            }
        )
    elif _type == "Article":
        issn = py_.get(meta, "isPartOf.issn")
        url = py_.get(meta, "publisher.url")
        container = compact(
            {
                "type": "Periodical",
                "title": py_.get(meta, "isPartOf.name"),
                "identifier": issn
                if issn is not None
                else url
                if url is not None
                else None,
                "identifierType": "ISSN"
                if issn is not None
                else "URL"
                if url is not None
                else None,
            }
        )
    else:
        container = {}

    references = wrap(schema_org_references(meta))
    funding_references = [
        get_funding_reference(i) for i in wrap(meta.get("funder", None))
    ]

    descriptions = [
        {
            "description": sanitize(i),
            "descriptionType": "Abstract",
        }
        for i in wrap(meta.get("description"))
    ]

    # convert keywords as comma-separated string into list
    subj = meta.get("keywords", None)
    if isinstance(subj, str):
        subj = subj.lower().split(", ")
    subjects = [name_to_fos(i) for i in wrap(subj)]

    if isinstance(meta.get("inLanguage"), str):
        language = meta.get("inLanguage")
    elif isinstance(meta.get("inLanguage"), list):
        language = py_.get(meta, "inLanguage.0")
    elif isinstance(meta.get("inLanguage"), dict):
        language = py_.get(meta, "inLanguage.alternateName") or py_.get(
            meta, "inLanguage.name"
        )
    else:
        language = None

    geo_locations = [
        schema_org_geolocation(i) for i in wrap(meta.get("spatialCoverage", None))
    ]
    alternate_identifiers = None
    provider = (
        get_doi_ra(_id)
        if doi_from_url(_id)
        else parse_attributes(meta.get("provider", None), content="name", first=True)
    )
    state = None

    return {
        # required attributes
        "id": _id,
        "type": _type,
        "url": normalize_url(meta.get("url", None)),
        "contributors": presence(contributors),
        "titles": titles,
        "publisher": publisher,
        "date": compact(date),
        # recommended and optional attributes
        "additional_type": additional_type,
        "subjects": presence(subjects),
        "language": language,
        "alternate_identifiers": alternate_identifiers,
        "sizes": None,
        "formats": None,
        "version": meta.get("version", None),
        "license": license_,
        "descriptions": presence(descriptions),
        "geo_locations": presence(geo_locations),
        "funding_references": presence(funding_references),
        "references": presence(references),
        # optional attributes
        "container": container,
        "provider": provider,
        "state": state,
    } | read_options


def schema_org_related_item(meta, relation_type=None):
    """Related items"""
    normalize_ids(
        ids=wrap(meta.get(relation_type, None)),
        relation_type=SO_TO_DC_RELATION_TYPES.get(relation_type),
    )


def schema_org_reverse_related_item(meta, relation_type=None):
    """Reverse related items"""
    normalize_ids(
        ids=wrap(py_.get(meta, f"@reverse.{relation_type}")),
        relation_type=SO_TO_DC_REVERSE_RELATION_TYPES.get(relation_type),
    )


def schema_org_is_identical_to(meta):
    """isIdenticalTo is a special case because it can be a string or an object."""
    schema_org_related_item(meta, relation_type="sameAs")


def schema_org_is_part_of(meta):
    """isPartOf is a special case because it can be a string or an object."""
    schema_org_related_item(meta, relation_type="isPartOf")


def schema_org_has_part(meta):
    """hasPart is a special case because it can be a string or an object."""
    schema_org_related_item(meta, relation_type="hasPart")


def schema_org_is_previous_version_of(meta):
    """isPreviousVersionOf is a special case because it can be a string or an object."""
    schema_org_related_item(meta, relation_type="PredecessorOf")


def schema_org_is_new_version_of(meta):
    """isNewVersionOf is a special case because it can be a string or an object."""
    schema_org_related_item(meta, relation_type="SuccessorOf")


def schema_org_references(meta):
    """references is a special case because it can be a string or an object."""
    schema_org_related_item(meta, relation_type="citation")


def schema_org_is_referenced_by(meta):
    """isReferencedBy is a special case because it can be a string or an object."""
    schema_org_reverse_related_item(meta, relation_type="citation")


def schema_org_is_supplement_to(meta):
    """isSupplementTo is a special case because it can be a string or an object."""
    schema_org_reverse_related_item(meta, relation_type="isBasedOn")


def schema_org_is_supplemented_by(meta):
    """isSupplementedBy is a special case because it can be a string or an object."""
    schema_org_related_item(meta, relation_type="isBasedOn")


def schema_org_geolocation(geo_location: Optional[dict]) -> Optional[dict]:
    """Geolocations in Schema.org format"""
    if not isinstance(geo_location, dict):
        return None

    _type = py_.get(geo_location, "geo.@type")
    longitude = py_.get(geo_location, "geo.longitude")
    latitude = py_.get(geo_location, "geo.latitude")

    if _type == "GeoCoordinates":
        return {
            "geoLocationPoint": {"pointLongitude": longitude, "pointLatitude": latitude}
        }
    return None


def get_html_meta(soup):
    """Get metadata from HTML meta tags"""
    data = {}
    pid = (
        soup.select_one("meta[name='citation_doi']")
        or soup.select_one("meta[name='dc.identifier']")
        or soup.select_one('[rel="canonical"]')
    )
    if pid is not None:
        data["@id"] = pid.get("content", None) or pid.get("href", None)

    _type = soup.select_one("meta[property='og:type']")
    data["@type"] = _type["content"].capitalize() if _type else None

    url = soup.select_one("meta[property='og:url']")
    data["url"] = url["content"] if url else None

    title = (
        soup.select_one("meta[name='citation_title']")
        or soup.select_one("meta[name='dc.title']")
        or soup.select_one("meta[property='og:title']")
    )
    data["name"] = title["content"] if title else None

    description = soup.select_one("meta[name='citation_abstract']") or soup.select_one(
        "meta[name='dc.description']"
    )
    data["description"] = description["content"] if description else None

    keywords = soup.select_one("meta[name='citation_keywords']")
    data["keywords"] = (
        str(keywords["content"]).replace(";", ",").rstrip(", ") if keywords else None
    )

    date_published = soup.select_one(
        "meta[name='citation_publication_date']"
    ) or soup.select_one("meta[name='dc.date']")
    data["datePublished"] = (
        get_iso8601_date(date_published["content"]) if date_published else None
    )

    license_ = soup.select_one("meta[name='dc.rights']")
    data["license"] = license_["content"] if license_ else None

    lang = soup.select_one("meta[name='dc.language']") or soup.select_one(
        "meta[name='citation_language']"
    )
    if lang is not None:
        data["inLanguage"] = lang["content"]
    else:
        lang = soup.select_one("html")["lang"]
        if lang is not None:
            data["inLanguage"] = lang

    publisher = soup.select_one("meta[property='og:site_name']")
    data["publisher"] = {"name": publisher["content"]} if publisher else None

    name = soup.select_one("meta[property='og:site_name']")
    issn = soup.select_one("meta[name='citation_issn']")
    data["isPartOf"] = compact(
        {
            "name": name["content"] if name else None,
            "issn": issn["content"] if issn else None,
        }
    )
    return data


def get_funding_reference(dct):
    """Get funding reference"""
    return compact(
        {
            "funderName": dct.get("name", None),
            "funderIdentifier": dct.get("@id", None),
            "funderIdentifierType": "Crossref Funder ID"
            if dct.get("@id", None)
            else None,
        }
    )



"""kbase reader for Commonmeta"""
from pydash import py_

from ..utils import normalize_url, normalize_doi, from_curie, from_kbase
from ..base_utils import compact, wrap, presence, sanitize
from ..author_utils import get_authors
from ..date_utils import normalize_date_dict
from ..doi_utils import doi_from_url, validate_doi
from ..constants import (
    COMMONMETA_RELATION_TYPES,
    Commonmeta,
)


def read_kbase(data: dict, **kwargs) -> Commonmeta:
    """read_kbase"""
    meta = data.get("credit_metadata", {})
    read_options = kwargs or {}

    _id = from_curie(meta.get("identifier", None))
    _type = "Dataset"
    contributors = get_authors(from_kbase(wrap(meta.get("contributors", None))))

    publisher = meta.get("publisher", None)
    if publisher is not None:
        publisher = {
            "id": from_curie(publisher.get("organization_id", None)),
            "name": publisher.get("organization_name", None),
        }
    titles = [format_title(i) for i in wrap(meta.get("titles", None))]

    date: dict = {}

    # convert date list to dict
    for sub in wrap(meta.get("dates", None)):
        data_type = sub.get("event", None)
        date[data_type.capitalize() if data_type else None] = sub.get("date", None)
    date = normalize_date_dict(date)

    container = compact(
        {
            "id": "https://www.re3data.org/repository/r3d100012864",
            "type": "DataRepository",
            "title": "KBase",
        }
    )
    license_ = meta.get("license", None)
    if license_:
        license_ = license_[0]
    descriptions = meta.get("descriptions", None)
    for des in wrap(descriptions):
        des["description"] = sanitize(des["description_text"])
        des["type"] = (
            des["description_type"]
            if des["description_type"] in ["Abstract", "Description", "Summary"]
            else None
        )
        py_.omit(des, ["description_text", "description_type"])
    language = meta.get("language", None)

    # subjects = [name_to_fos(i) for i in wrap(py_.get(meta, "metadata.keywords"))]

    version = meta.get("version", None)
    references = get_references(wrap(meta.get("related_identifiers")))
    related_identifiers = get_related_identifiers(wrap(meta.get("related_identifiers")))
    funding_references = get_funding_references(wrap(meta.get("funding", None)))
    files = [get_file(i) for i in wrap(meta.get("content_url"))]

    state = "findable" if meta or read_options else "not_found"

    return {
        # required properties
        "id": _id,
        "type": _type,
        "doi": doi_from_url(_id),
        "url": normalize_url(meta.get("url", None)),
        "contributors": presence(contributors),
        "titles": titles,
        "publisher": publisher,
        "date": compact(date),
        # recommended and optional properties
        "additional_type": None,
        "subjects": None,
        "language": language,
        "alternate_identifiers": None,
        "sizes": None,
        "formats": None,
        "version": py_.get(meta, "metadata.version"),
        "license": presence(license_),
        "descriptions": descriptions,
        "geo_locations": None,
        "funding_references": presence(funding_references),
        "references": presence(references),
        "related_identifiers": presence(related_identifiers),
        # other properties
        "files": presence(files),
        "container": container,
        "provider": "KBase",
        "state": state,
        "schema_version": py_.get(data, "credit_metadata_schema_version"),
    } | read_options


def format_title(title: dict) -> dict:
    """format_title"""
    _type = title.get("title_type", None)
    return compact(
        {
            "title": title.get("title", None),
            "type": _type
            if _type in ["AlternativeTitle", "Subtitle", "TranslatedTitle"]
            else None,
        }
    )


def get_references(references: list) -> list:
    """get_references"""

    def is_reference(reference):
        """is_reference"""
        return reference.get("relationship_type", None) in [
            "DataCite:Cites",
            "DataCite:References",
            "DataCite:IsSupplementedBy",
        ]

    def map_reference(reference):
        """map_reference"""
        identifier = from_curie(reference.get("id", None))
        identifier_type = "DOI" if validate_doi(identifier) else "URL"
        if identifier and identifier_type == "DOI":
            reference["doi"] = normalize_doi(identifier)
        elif identifier and identifier_type == "URL":
            reference["url"] = normalize_url(identifier)
        reference = py_.omit(
            reference,
            [
                "id",
                "relationship_type",
            ],
        )
        return reference

    return [map_reference(i) for i in references if is_reference(i)]


def get_file(file: str) -> dict:
    """get_file"""
    return compact({"url": file})


def get_related_identifiers(related_identifiers: list) -> list:
    """get_related_identifiers"""

    def map_related_identifier(related_identifier: dict) -> dict:
        """get_related_identifier"""
        identifier = from_curie(related_identifier.get("id", None))
        _type = related_identifier.get("relationship_type", None)
        # remove DataCite: and Crossref: prefixes
        _type = _type.split(":")[1] if _type else None
        if normalize_url(identifier):
            identifier = normalize_url(identifier)
        # TODO: resolvable url for other identifier types
        else:
            identifier = None
        return {
            "id": identifier,
            "type": _type,
        }

    identifiers = [map_related_identifier(i) for i in related_identifiers]
    return [i for i in identifiers if i["type"] in COMMONMETA_RELATION_TYPES]


def get_funding_references(funding_references: list) -> list:
    """get_funding_references"""

    def map_funding_reference(funding_reference: dict) -> dict:
        """map_funding_reference"""
        funder_identifier = py_.get(funding_reference, "funder.organization_id", None)
        funder_identifier_type = (
            funder_identifier.split(":")[0] if funder_identifier else None
        )
        return compact(
            {
                "funderIdentifier": from_curie(funder_identifier),
                "funderIdentifierType": funder_identifier_type,
                "funderName": py_.get(
                    funding_reference, "funder.organization_name", None
                ),
                "awardNumber": funding_reference.get("grant_id", None),
                "award_uri": funding_reference.get("grant_url", None),
            }
        )

    return [map_funding_reference(i) for i in funding_references]


def format_descriptions(descriptions: list) -> list:
    """format_descriptions"""
    return [
        {
            "description": sanitize(i),
            "descriptionType": "Abstract" if index == 0 else "Other",
        }
        for index, i in enumerate(descriptions)
        if i
    ]



"""Base utilities for commonmeta-py"""
import html
from os import path
import re
import xmltodict
from typing import Optional, Union
import nh3


def wrap(item) -> list:
    """Turn None, dict, or list into list"""
    if item is None:
        return []
    if isinstance(item, list):
        return item
    return [item]


def unwrap(lst: list) -> Optional[Union[dict, list]]:
    """Turn list into dict or None, depending on list size"""
    if len(lst) == 0:
        return None
    if len(lst) == 1:
        return lst[0]
    return lst


def presence(
    item: Optional[Union[dict, list, str]],
) -> Optional[Union[dict, list, str]]:
    """Turn empty list, dict or str into None"""
    return None if item is None or len(item) == 0 or item == [{}] else item


def compact(dict_or_list: Union[dict, list]) -> Optional[Union[dict, list]]:
    """Remove None from dict or list"""
    if isinstance(dict_or_list, dict):
        return {k: v for k, v in dict_or_list.items() if v is not None}
    if isinstance(dict_or_list, list):
        lst = [compact(i) for i in dict_or_list]
        return lst if len(lst) > 0 else None

    return None


def parse_attributes(
    element: Union[str, dict, list], **kwargs
) -> Optional[Union[str, list]]:
    """extract attributes from a string, dict or list"""

    def parse_item(item):
        if isinstance(item, dict):
            return item.get(html.unescape(content), None)
        return html.unescape(item)

    content = kwargs.get("content", "#text")
    if isinstance(element, str) and kwargs.get("content", None) is None:
        return html.unescape(element)
    if isinstance(element, dict):
        return element.get(html.unescape(content), None)
    if isinstance(element, list):
        arr = [parse_item(i) for i in element if i]
        arr = arr[0] if len(arr) > 0 and kwargs.get("first") else unwrap(arr)
        return arr


def parse_xml(string: Optional[str], **kwargs) -> Optional[Union[dict, list]]:
    """Parse XML into dict. Set default options, and options for Crossref XML"""
    if string is None:
        return None
    if path.exists(string):
        with open(string, encoding="utf-8") as file:
            string = file.read()

    if kwargs.get("dialect", None) == "crossref":
        # remove namespaces from xml
        namespaces = {
            "http://www.crossref.org/schema/5.3.1": None,
            "http://www.crossref.org/qrschema/3.0": None,
            "http://www.crossref.org/xschema/1.0": None,
            "http://www.crossref.org/xschema/1.1": None,
            "http://www.crossref.org/AccessIndicators.xsd": None,
            "http://www.crossref.org/fundref.xsd": None,
            "http://www.ncbi.nlm.nih.gov/JATS1": None,
        }

        kwargs["process_namespaces"] = True
        kwargs["namespaces"] = namespaces
        kwargs["force_list"] = {
            "person_name",
            "organization",
            "titles",
            "item",
            "citation",
        }

    kwargs["attr_prefix"] = ""
    kwargs["dict_constructor"] = dict
    kwargs.pop("dialect", None)
    return xmltodict.parse(string, **kwargs)


def sanitize(text: str, **kwargs) -> str:
    """Sanitize text"""
    # default whitelisted HTML tags
    tags = kwargs.get("tags", None) or {
        "b",
        "br",
        "code",
        "em",
        "i",
        "sub",
        "sup",
        "strong",
    }
    attributes = kwargs.get("attributes", None)
    string = nh3.clean(text, tags=tags, attributes=attributes, link_rel=None)
    # remove excessive internal whitespace
    return " ".join(re.split(r"\s+", string, flags=re.UNICODE))



"""codemeta reader for commonmeta-py"""
from typing import Optional
from collections import defaultdict
import httpx

from ..utils import (
    normalize_id,
    from_schema_org_creators,
    name_to_fos,
    dict_to_spdx,
    github_as_codemeta_url,
    github_as_repo_url,
    doi_from_url,
)
from ..base_utils import wrap, presence, compact, sanitize
from ..author_utils import get_authors
from ..constants import (
    Commonmeta,
    SO_TO_CM_TRANSLATIONS,
)


def get_codemeta(pid: str, **kwargs) -> dict:
    """get_codemeta"""
    url = str(github_as_codemeta_url(pid))
    response = httpx.get(url, timeout=10, **kwargs)
    if response.status_code != 200:
        return {"state": "not_found"}
    data = response.json()
    if data.get("codeRepository", None) is None:
        data["codeRepository"] = github_as_repo_url(url)

    return data


def read_codemeta(data: Optional[dict], **kwargs) -> Commonmeta:
    """read_codemeta"""
    if data is None:
        return {"state": "not_found"}
    meta = data

    read_options = kwargs or {}
    # ActiveSupport: : HashWithIndifferentAccess.new(options.except(: doi, : id, : url,
    # : sandbox, : validate, : ra)

    _id = normalize_id(meta.get("id", None) or meta.get("identifier", None))
    # id = normalize_id(options[:doi] | | meta.get('@id', None) | | meta.get('identifier', None))
    _type = SO_TO_CM_TRANSLATIONS.get(meta.get("@type", "Software"))
    # identifiers = Array.wrap(meta.get('identifier', None)).map do | r|
    #   r = normalize_id(r) if r.is_a?(String)
    #   if r.is_a?(String) & & URI(r) != 'doi.org'
    #     {'identifierType': 'URL', 'identifier': r}
    #   elsif r.is_a?(Hash)
    #     {'identifierType': get_identifier_type(
    #         r['propertyID']), 'identifier': r['value']}
    #   end
    # end.compact.uniq

    has_agents = meta.get("agents", None)
    authors = meta.get("authors", None) if has_agents is None else has_agents
    contributors = get_authors(from_schema_org_creators(wrap(authors)))
    contrib = get_authors(from_schema_org_creators(wrap(meta.get("editor", None))))
    if contrib:
        contributors += contrib
    date: dict = defaultdict(list)
    date["created"] = meta.get("dateCreated", None)
    date["published"] = meta.get("datePublished", None)
    date["updated"] = meta.get("dateModified", None)

    publisher = {"name": meta.get("publisher", None)}

    if meta.get("description", None):
        descriptions = [
            {
                "description": sanitize(str(meta.get("description"))),
                "descriptionType": "Abstract",
            }
        ]
    else:
        descriptions = None

    subjects = [name_to_fos(i) for i in wrap(meta.get("keywords", None))]

    has_title = meta.get("title", None)
    if has_title is None:
        titles = [{"title": meta.get("name", None)}]
    else:
        titles = [{"title": has_title}]

    license_ = meta.get("licenseId", None)
    if license_:
        license_ = dict_to_spdx({"id": meta.get("licenseId")})

    provider = "DataCite" if doi_from_url(_id) else "GitHub"
    state = "findable" if meta or read_options else "not_found"

    return {
        "id": _id,
        "type": _type,
        "url": normalize_id(meta.get("codeRepository", None)),
        "identifiers": None,
        "titles": titles,
        "contributors": presence(contributors),
        "publisher": publisher,
        "date": compact(date),
        "descriptions": descriptions,
        "license": license_,
        "version": meta.get("version", None),
        "subjects": presence(subjects),
        "provider": provider,
        "state": state,
    } | read_options



"""CSL-JSON writer for commonmeta-py"""
import orjson as json
from typing import Optional

from ..utils import pages_as_string, to_csl
from ..base_utils import wrap, presence, parse_attributes, compact
from ..date_utils import get_date_parts
from ..doi_utils import doi_from_url
from ..constants import CM_TO_CSL_TRANSLATIONS, Commonmeta


def write_csl(metadata: Commonmeta) -> Optional[str]:
    """Write CSL-JSON"""
    item = write_csl_item(metadata)
    if item is None:
        return None
    return json.dumps(item)


def write_csl_item(metadata) -> Optional[dict]:
    """Write CSL-JSON item"""
    if metadata is None or metadata.write_errors is not None:
        return None
    if len(wrap(metadata.contributors)) == 0:
        author = None
    else:
        author = to_csl(wrap(metadata.contributors))

    if metadata.type == "Software" and metadata.version is not None:
        _type = "book"
    else:
        _type = CM_TO_CSL_TRANSLATIONS.get(metadata.type, "Document")

    container = metadata.container or {}
    publisher = metadata.publisher or {}
    return compact(
        {
            "type": _type,
            "id": metadata.id,
            "DOI": doi_from_url(metadata.id),
            "URL": metadata.url,
            "categories": presence(
                parse_attributes(
                    wrap(metadata.subjects), content="subject", first=False
                )
            ),
            "language": metadata.language,
            "author": author,
            # "contributor": to_csl(wrap(metadata.contributors)),
            "issued": get_date_parts(metadata.date.get("published", None)),
            "submitted": get_date_parts(metadata.date.get("submitted"))
            if metadata.date.get("submitted", None)
            else None,
            "abstract": parse_attributes(
                metadata.descriptions, content="description", first=True
            ),
            "container-title": container.get("title", None),
            "volume": container.get("volume", None),
            "issue": container.get("issue", None),
            "page": pages_as_string(container),
            "publisher": publisher.get("name", None),
            "title": parse_attributes(metadata.titles, content="title", first=True),
            "copyright": metadata.license.get("id", None) if metadata.license else None,
            "version": metadata.version,
        }
    )


def write_csl_list(metalist):
    """Write CSL-JSON list"""
    if metalist is None:
        return None
    items = [write_csl_item(item) for item in metalist.items]
    return json.dumps(items)



"""datacite_xml reader for Commonmeta"""
from collections import defaultdict
import httpx
from pydash import py_

from ..base_utils import compact, wrap, presence, sanitize, parse_attributes
from ..author_utils import get_authors
from ..date_utils import strip_milliseconds, normalize_date_dict
from ..doi_utils import doi_from_url, doi_as_url, datacite_api_url, normalize_doi
from ..utils import normalize_url, normalize_cc_url, dict_to_spdx
from ..constants import DC_TO_CM_TRANSLATIONS, Commonmeta


def get_datacite_xml(pid: str, **kwargs) -> dict:
    """get_datacite_xml"""
    doi = doi_from_url(pid)
    if doi is None:
        return {"state": "not_found"}
    url = datacite_api_url(doi)
    response = httpx.get(url, timeout=10, **kwargs)
    if response.status_code != 200:
        return {"state": "not_found"}
    return py_.get(response.json(), "data.attributes", {})


def read_datacite_xml(data: dict, **kwargs) -> Commonmeta:
    """read_datacite_xml"""
    if data is None:
        return {"state": "not_found"}

    read_options = kwargs or {}

    meta = data.get("resource", {})

    doi = parse_attributes(meta.get("identifier", None))
    _id = doi_as_url(doi) if doi else None
    #         identifiers = Array.wrap(meta.dig('alternateIdentifiers', 'alternateIdentifier')).map do |r|
    #           if r['__content__'].present?
    #             { 'identifierType' => get_identifier_type(r['alternateIdentifierType']),
    #               'identifier' => r['__content__'] }
    #           end
    #         end.compact

    resource__typegeneral = py_.get(meta, "resourceType.resourceTypeGeneral")
    _type = DC_TO_CM_TRANSLATIONS.get(resource__typegeneral, "Other")
    additional_type = py_.get(meta, "resourceType.#text")

    def format_title(title):
        """format_title"""
        if isinstance(title, str):
            return {"title": title}
        if isinstance(title, dict):
            return {
                "title": title.get("#text", None),
                "titleType": title.get("titleType", None),
                "lang": title.get("xml:lang", None),
            }
        return None

    titles = [format_title(i) for i in wrap(py_.get(meta, "titles.title"))]

    contributors = get_authors(wrap(py_.get(meta, "creators.creator")))
    contrib = get_authors(wrap(meta.get("contributors", None)))
    if contrib:
        contributors = contributors + contrib
    publisher = {"name": py_.get(meta, "publisher")}
    date = get_dates(
        wrap(py_.get(meta, "dates.date")), meta.get("publicationYear", None)
    )

    def format_description(description):
        """format_description"""
        if isinstance(description, str):
            return {"description": description, "descriptionType": "Abstract"}
        if isinstance(description, dict):
            return compact(
                {
                    "description": sanitize(description.get("#text", None)),
                    "descriptionType": description.get("descriptionType", "Abstract"),
                    "lang": description.get("xml:lang", None),
                }
            )
        return None

    descriptions = [
        format_description(i) for i in wrap(py_.get(meta, "descriptions.description"))
    ]

    def format_subject(subject):
        """format_subject"""
        if isinstance(subject, str):
            return {"subject": subject, "subjectScheme": "None"}
        if isinstance(subject, dict):
            return compact(
                {
                    "subject": subject.get("#text", None),
                    "subjectScheme": subject.get("subjectScheme", None),
                    "lang": subject.get("xml:lang", None),
                }
            )
        return None

    subjects = [format_subject(i) for i in wrap(py_.get(meta, "subjects.subject")) if i]

    def format_geo_location(geo_location):
        """format_geo_location"""
        if isinstance(geo_location, str):
            return {"geoLocationPlace": geo_location}
        if isinstance(geo_location, dict):
            return compact(
                {
                    "geoLocationPoint": compact(
                        {
                            "pointLatitude": compact(
                                geo_location.get("geoLocationPoint.pointLatitude", None)
                            ),
                            "pointLongitude": compact(
                                geo_location.get(
                                    "geoLocationPoint.pointLongitude", None
                                )
                            ),
                        }
                    ),
                    "geoLocationBox": compact(
                        {
                            "westBoundLongitude": compact(
                                geo_location.get(
                                    "geoLocationBox.westBoundLongitude", None
                                )
                            ),
                            "eastBoundLongitude": compact(
                                geo_location.get(
                                    "geoLocationBox.eastBoundLongitude", None
                                )
                            ),
                            "southBoundLatitude": compact(
                                geo_location.get(
                                    "geoLocationBox.southBoundLatitude", None
                                )
                            ),
                            "northBoundLatitude": compact(
                                geo_location.get(
                                    "geoLocationBox.northBoundLatitude", None
                                )
                            ),
                        }
                    ),
                    "geoLocationPolygon": {
                        "polygonPoint": compact(
                            {
                                "pointLatitude": geo_location.get(
                                    "geoLocationPolygon.polygonPoint.pointLatitude",
                                    None,
                                ),
                                "pointLongitude": geo_location.get(
                                    "geoLocationPolygon.polygonPoint.pointLongitude",
                                    None,
                                ),
                            }
                        )
                    },
                    "geoLocationPlace": geo_location.get("geoLocationPlace", None),
                }
            )
        return None

    geo_locations = []  # [format_geo_location(i) for i in wrap(py_.get(meta, "geoLocations.geoLocation")) if i]

    def map_size(size):
        """map_size"""
        return size.get("#text")

    sizes = [map_size(i) for i in wrap(meta.get("sizes", None))]

    def map_format(format_):
        """map_format"""
        return format_.get("#text")

    formats = [map_format(i) for i in wrap(meta.get("formats", None))]

    def map_rights(rights):
        """map_rights"""
        return compact(
            {
                "rights": rights.get("#text", None),
                "url": rights.get("rightsURI", None),
                "lang": rights.get("xml:lang", None),
            }
        )

    license_ = wrap(py_.get(meta, "rightsList.rights"))
    if len(license_) > 0:
        license_ = normalize_cc_url(license_[0].get("rightsURI", None))
        license_ = dict_to_spdx({"url": license_}) if license_ else None

    references = get_xml_references(
        wrap(py_.get(meta, "relatedIdentifiers.relatedIdentifier"))
    )

    def map_funding_reference(funding_reference):
        """map_funding_reference"""
        return {
            "funderName": funding_reference.get("funderName", None),
            "funderIdentifier": funding_reference.get("funderIdentifier", None),
            "funderIdentifierType": funding_reference.get("funderIdentifierType", None),
            "awardNumber": funding_reference.get("awardNumber", None),
            "awardTitle": funding_reference.get("awardTitle", None),
        }

    funding_references = []  # [map_funding_reference(i) for i in wrap(py_.get(meta, "fundingReferences.fundingReference"))]

    files = meta.get("contentUrl", None)
    state = "findable" if _id or read_options else "not_found"

    return {
        # required properties
        "id": _id,
        "type": _type,
        "doi": doi_from_url(_id),
        "url": normalize_url(meta.get("url", None)),
        "contributors": presence(contributors),
        "titles": compact(titles),
        "publisher": publisher,
        "date": date,
        # recommended and optional properties
        "additional_type": presence(additional_type),
        "subjects": presence(subjects),
        "language": meta.get("language", None),
        "alternate_identifiers": presence(meta.get("alternateIdentifiers", None)),
        "sizes": presence(sizes),
        "formats": presence(formats),
        "version": meta.get("version", None),
        "license": presence(license_),
        "descriptions": presence(descriptions),
        "geo_locations": presence(geo_locations),
        "funding_references": presence(funding_references),
        "references": presence(references),
        # other properties
        "date_created": strip_milliseconds(meta.get("created", None)),
        "date_registered": strip_milliseconds(meta.get("registered", None)),
        "date_published": strip_milliseconds(meta.get("published", None)),
        "date_updated": strip_milliseconds(meta.get("updated", None)),
        "files": presence(files),
        "container": presence(meta.get("container", None)),
        "provider": "DataCite",
        "state": state,
        "schema_version": meta.get("xmlns", None),
    } | read_options


def get_xml_references(references: list) -> list:
    """get_xml_references"""

    def is_reference(reference):
        """is_reference"""
        return reference.get("relationType", None) in [
            "Cites",
            "References",
        ] and reference.get("relatedIdentifierType", None) in ["DOI", "URL"]

    def map_reference(reference):
        """map_reference"""
        identifier = reference.get("relatedIdentifier", None)
        identifier_type = reference.get("relatedIdentifierType", None)
        if identifier and identifier_type == "DOI":
            reference["doi"] = normalize_doi(identifier)
        elif identifier and identifier_type == "URL":
            reference["url"] = normalize_url(identifier)
        reference = py_.omit(
            reference,
            [
                "relationType",
                "relatedIdentifier",
                "relatedIdentifierType",
                "resourceTypeGeneral",
                "schemeType",
                "schemeUri",
                "relatedMetadataScheme",
            ],
        )
        return reference

    return [map_reference(i) for i in references if is_reference(i)]


def get_dates(dates: list, publication_year) -> dict:
    """convert date list to dict, rename and/or remove some keys"""
    date: dict = defaultdict(list)
    for sub in dates:
        date[sub.get("dateType", None)] = sub.get("#text", None)
    if date.get("Issued", None) is None and publication_year is not None:
        date["Issued"] = str(publication_year)
    return normalize_date_dict(date)



"""JSON Feed reader for commonmeta-py"""
from typing import Optional
import httpx
from pydash import py_

from ..utils import (
    compact,
    normalize_url,
    from_json_feed,
    wrap,
    dict_to_spdx,
    name_to_fos,
    validate_url,
    encode_doi,
)
from ..author_utils import get_authors
from ..base_utils import presence, sanitize, parse_attributes
from ..date_utils import get_date_from_unix_timestamp
from ..doi_utils import (
    normalize_doi,
    validate_prefix,
    validate_doi,
    doi_from_url,
    is_rogue_scholar_doi,
)
from ..constants import Commonmeta


def get_json_feed_item(pid: str, **kwargs) -> dict:
    """get_json_feed_item"""
    if pid is None:
        return {"state": "not_found"}
    url = normalize_url(pid)
    response = httpx.get(url, timeout=10, follow_redirects=True, **kwargs)
    if response.status_code != 200:
        return {"state": "not_found"}
    return response.json()


def read_json_feed_item(data: Optional[dict], **kwargs) -> Commonmeta:
    """read_json_feed_item"""
    if data is None:
        return {"state": "not_found"}
    meta = data
    read_options = kwargs or {}

    url = normalize_url(meta.get("url", None))
    _id = normalize_doi(read_options.get("doi", None) or meta.get("doi", None)) or url
    _type = "Article"

    # optionally generate a DOI if missing but a DOI prefix is provided
    prefix = read_options.get("prefix", None) or py_.get(meta, "blog.prefix", None)
    if doi_from_url(_id) is None and prefix is not None:
        _id = encode_doi(prefix)

    if meta.get("authors", None):
        contributors = get_authors(from_json_feed(wrap(meta.get("authors"))))
    else:
        contributors = [{"type": "Organization", "name": ":(unav)"}]

    title = parse_attributes(meta.get("title", None))
    titles = [{"title": sanitize(title)}] if title else None

    publisher = py_.get(meta, "blog.title", None)
    if publisher is not None:
        publisher = {"name": publisher}

    date: dict = {}
    date["published"] = (
        get_date_from_unix_timestamp(meta.get("published_at", None))
        if meta.get("published_at", None)
        else None
    )
    date["updated"] = (
        get_date_from_unix_timestamp(meta.get("updated_at", None))
        if meta.get("updated_at", None)
        else None
    )

    license_ = py_.get(meta, "blog.license", None)
    if license_ is not None:
        license_ = dict_to_spdx({"url": license_})

    container = compact(
        {
            "type": "Periodical",
            "title": py_.get(meta, "blog.title", None),
            "identifier": py_.get(meta, "blog.issn", None),
            "identifierType": "ISSN" if py_.get(meta, "blog.issn", None) else None,
        }
    )

    description = meta.get("summary", None)
    if description is not None:
        descriptions = [
            {"description": sanitize(description), "descriptionType": "Abstract"}
        ]
    else:
        descriptions = None
    category = py_.get(meta, "blog.category", None)
    if category is not None:
        subjects = [name_to_fos(py_.human_case(category))]
    references = get_references(wrap(meta.get("reference", None)))
    funding_references = get_funding_references(meta)
    related_identifiers = get_related_identifiers(wrap(meta.get("relationships", None)))
    alternate_identifiers = [
        {"alternateIdentifier": meta.get("id"), "alternateIdentifierType": "UUID"}
    ]
    files = get_files(_id)
    state = "findable" if meta or read_options else "not_found"

    return {
        # required properties
        "id": _id,
        "type": _type,
        "url": url,
        "contributors": presence(contributors),
        "titles": presence(titles),
        "publisher": publisher,
        "date": compact(date),
        # recommended and optional properties
        "subjects": presence(subjects),
        "language": meta.get("language", None),
        "alternate_identifiers": alternate_identifiers,
        "sizes": None,
        "formats": None,
        "version": None,
        "license": license_,
        "descriptions": descriptions,
        "geo_locations": None,
        "funding_references": presence(funding_references),
        "references": references,
        "related_identifiers": presence(related_identifiers),
        "files": files,
        # other properties
        "container": presence(container),
        # "provider": get_doi_ra(_id),
        "state": state,
        "schema_version": None,
    } | read_options


def get_references(references: list) -> list:
    """get json feed references."""

    def get_reference(reference: dict) -> Optional[dict]:
        if reference is None or not isinstance(reference, dict):
            return None
        try:
            if reference.get("doi", None) and validate_doi(reference.get("doi")):
                doi = normalize_doi(reference.get("doi"))
                return compact(
                    {
                        "doi": doi,
                        "title": reference.get("title", None),
                        "publicationYear": reference.get("publicationYear", None),
                        "url": reference.get("url", None),
                    }
                )

            elif (
                reference.get("url", None)
                and validate_url(reference.get("url")) == "URL"
            ):
                response = httpx.head(reference.get("url", None), timeout=10)
                # check that URL resolves.
                # TODO: check for redirects
                if response.status_code in [404]:
                    return None
                return {
                    "url": reference.get("url"),
                }
        except Exception as error:
            print(error)
            return None

    def number_reference(reference: dict, index: int) -> dict:
        """number reference"""
        reference["key"] = f"ref{index +1}"
        return reference

    references = [get_reference(i) for i in references]
    return [
        number_reference(i, index)
        for index, i in enumerate(references)
        if i is not None
    ]


def get_funding_references(meta: Optional[dict]) -> Optional[list]:
    """get json feed funding references.
    Check that relationships resolve and have type "HasAward" or
    funding is provided by blog metadata"""

    if meta is None or not isinstance(meta, dict):
        return None

    def format_funding(url: str) -> dict:
        """format funding. Prefix 10.3030 means funder is European Commission"""
        return {
            "funderName": "European Commission",
            "funderIdentifier": "https://doi.org/10.13039/501100000780",
            "funderIdentifierType": "Crossref Funder ID",
            "awardNumber": url.split("/")[-1],
        }

    awards = [
        format_funding(i.get("url"))
        for i in wrap(meta.get("relationships", None))
        if i.get("type", None) == "HasAward"
        and validate_prefix(i.get("url", None)) == "10.3030"
    ]
    funding = py_.get(meta, "blog.funding", None)
    if funding is not None:
        awards += [
            {
                "funderName": funding.get("funder_name", None),
                "funderIdentifier": funding.get("funder_id", None),
                "funderIdentifierType": "Crossref Funder ID",
                "awardTitle": funding.get("award", None),
                "awardNumber": funding.get("award_number", None),
            }
        ]
    return awards


def get_related_identifiers(relationships: Optional[list]) -> Optional[list]:
    """get json feed related identifiers.
    Check that relationships resolve and has a supported type"""
    supported_types = ["IsIdenticalTo", "IsPreprintOf", "IsTranslationOf"]

    def format_relationship(relationship: dict) -> dict:
        """format relationship"""
        return {
            "id": relationship.get("url", None),
            "type": relationship.get("type", None),
        }

    return [
        format_relationship(i)
        for i in relationships
        if i.get("type", None) in supported_types
    ]


def get_files(pid: str) -> Optional[list]:
    """get json feed file links"""
    doi = doi_from_url(pid)
    if not is_rogue_scholar_doi(doi):
        return None
    return [
        {
            "mimeType": "text/markdown",
            "url": f"https://api.rogue-scholar.org/posts/{doi}.md",
        },
        {
            "mimeType": "application/pdf",
            "url": f"https://api.rogue-scholar.org/posts/{doi}.pdf",
        },
        {
            "mimeType": "application/epub+zip",
            "url": f"https://api.rogue-scholar.org/posts/{doi}.epub",
        },
        {
            "mimeType": "application/xml",
            "url": f"https://api.rogue-scholar.org/posts/{doi}.xml",
        },
    ]


def get_json_feed_item_uuid(id: str):
    """get JSON Feed item by uuid"""
    if id is None:
        return None
    url = f"https://api.rogue-scholar.org/posts/{id}"
    response = httpx.get(url, timeout=10)
    if response.status_code != 200:
        return response.json()
    post = response.json()
    return py_.pick(
        post,
        [
            "id",
            "guid",
            "url",
            "doi",
            "title",
            "blog.slug",
            "blog.issn",
            "blog.prefix",
            "blog.status",
            "published_at",
            "updated_at",
            "indexed_at",
        ],
    )


def get_json_feed_blog_slug(id: str):
    """get JSON Feed item by id and return blog slug"""
    if id is None:
        return None
    url = f"https://api.rogue-scholar.org/posts/{id}"
    response = httpx.get(url, timeout=10)
    if response.status_code != 200:
        return response.json()
    post = response.json()
    return py_.get(post, "blog.slug", None)


def get_json_feed_blog_slug(id: str):
    """get JSON Feed item by id and return blog slug"""
    if id is None:
        return None
    url = f"https://api.rogue-scholar.org/posts/#{id}"
    response = httpx.get(url, timeout=10)
    if response.status_code != 200:
        return None
    post = response.json()
    return py_.get(post, "blog.slug", None)



"""cff reader for commonmeta-py"""
from typing import Optional
from urllib.parse import urlparse
import httpx
import yaml

from ..utils import (
    normalize_id,
    name_to_fos,
    dict_to_spdx,
    normalize_orcid,
    github_as_cff_url,
    github_as_repo_url,
)
from ..base_utils import compact, wrap, presence, sanitize, parse_attributes
from ..date_utils import get_iso8601_date
from ..constants import Commonmeta


def get_cff(pid: str, **kwargs) -> dict:
    """get_cff"""
    url = github_as_cff_url(pid)
    response = httpx.get(url, timeout=10, **kwargs)
    if response.status_code != 200:
        return {"state": "not_found"}
    text = response.text
    repo_url = github_as_repo_url(url)
    data = yaml.safe_load(text)

    # collect metadata not included in the CFF file
    if data.get("repository-code", None) is None:
        data["repository-code"] = repo_url

    return data


def read_cff(data: Optional[dict], **kwargs) -> Commonmeta:
    """read_cff"""
    if data is None:
        return {"state": "not_found"}
    meta = data

    read_options = kwargs or {}

    # read_options = ActiveSupport::HashWithIndifferentAccess.new(options.except(:doi, :id, :url, :sandbox, :validate, :ra))

    # identifiers = Array.wrap(meta.fetch('identifiers', nil)).map do |r|
    #   r = normalize_id(r) if r.is_a?(String)
    #   if r.is_a?(String) && URI(r).host != 'doi.org'
    #     { 'identifierType' => 'URL', 'identifier' => r }
    #   elsif r.is_a?(Hash)
    #     { 'identifierType' => get_identifier_type(r['propertyID']), 'identifier' => r['value'] }
    #   end
    # end.compact.uniq

    _id = normalize_id(kwargs.get("doi", None) or meta.get("doi", None))
    # Array.wrap(meta.fetch('identifiers', nil)).find do |i|
    #                                                     i['type'] == 'doi'
    #                                                   end.fetch('value', nil))
    _type = "Software"
    url = normalize_id(meta.get("repository-code", None))
    contributors = cff_contributors(wrap(meta.get("authors", None)))

    if meta.get("title", None):
        titles = [{"title": meta.get("title", None)}]
    else:
        titles = []

    date = {
        "published": get_iso8601_date(meta.get("date-released"))
        if meta.get("date-released", None)
        else None
    }

    publisher = (
        {"name": "GitHub"} if url and urlparse(url).hostname == "github.com" else None
    )

    if meta.get("abstract", None):
        descriptions = [
            {
                "description": sanitize(meta.get("abstract")),
                "descriptionType": "Abstract",
            }
        ]
    else:
        descriptions = []

    subjects = [name_to_fos(i) for i in wrap(meta.get("keywords", None))]

    license_ = meta.get("license", None)
    if license_ is not None:
        license_ = dict_to_spdx({"id": meta.get("license")})

    references = cff_references(wrap(meta.get("references", None)))

    state = "findable" if meta or read_options else "not_found"

    return {
        "id": _id,
        "type": _type,
        # 'identifiers' => identifiers,
        "url": url,
        "titles": titles,
        "contributors": presence(contributors),
        "publisher": publisher,
        "references": presence(references),
        "date": date,
        "descriptions": presence(descriptions),
        "license": license_,
        "version": meta.get("version", None),
        "subjects": presence(subjects),
        "provider": "DataCite" if _id else "GitHub",
        "state": state,
    } | read_options


def cff_contributors(contributors):
    """cff_contributors"""

    def format_affiliation(affiliation):
        """format_affiliation"""
        if isinstance(affiliation, str):
            return {"name": affiliation}
        if isinstance(affiliation, dict):
            return compact(affiliation)
        return None
        #         if a.is_a?(Hash)
        #   a
        # elsif a.is_a?(Hash) && a.key?('#text_') && a['#text'].strip.blank?
        #   nil
        # elsif a.is_a?(Hash) && a.key?('#text')
        #   { 'name' => a['#text'] }
        # elsif a.strip.blank

    def format_element(i):
        """format_element"""
        if normalize_orcid(parse_attributes(i.get("orcid", None))):
            _id = normalize_orcid(parse_attributes(i.get("orcid", None)))
        else:
            _id = None
        if i.get("given-names", None) or i.get("family-names", None) or _id:
            given_name = parse_attributes(i.get("given-names", None))
            family_name = parse_attributes(i.get("family-names", None))
            affiliation = compact(
                [format_affiliation(a) for a in wrap(i.get("affiliation", None))]
            )

            return compact(
                {
                    "id": _id,
                    "contributorRoles": ["Author"],
                    "type": "Person",
                    "givenName": given_name,
                    "familyName": family_name,
                    "affiliation": affiliation,
                }
            )
        return {
            "contributorRoles": ["Author"],
            "type": "Organization",
            "name": i.get("name", None) or i.get("#text", None),
        }

    return [format_element(i) for i in contributors]


def cff_references(references):
    """cff_references"""

    def is_reference(i):
        """is_reference"""
        return (
            next(
                (
                    item
                    for item in wrap(i.get("identifers", None))
                    if item.get("type", None) == "doi"
                ),
                None,
            )
            is not None
        )

    def map_reference(i):
        """map_element"""
        identifier = next(
            (
                item
                for item in wrap(i.get("identifers", None))
                if item.get("type", None) == "doi"
            ),
            None,
        )
        return compact(
            {"doi": normalize_id(parse_attributes(identifier.get("value", None)))}
        )

    return [map_reference(i) for i in references if is_reference(i)]



"""Commonmeta reader for commonmeta-py"""
from ..constants import Commonmeta


def read_commonmeta(data: dict, **kwargs) -> Commonmeta:
    """read_commonmeta"""
    if data is None:
        return {"state": "not_found"}
    meta = data

    read_options = kwargs or {}

    return meta | read_options



"""Citation writer for commonmeta-py"""
import orjson as json
import re
from pydash import py_
from citeproc import CitationStylesStyle, CitationStylesBibliography
from citeproc import Citation, CitationItem
from citeproc import formatter
from citeproc.source.json import CiteProcJSON
from citeproc_styles import get_style_filepath


def write_citation(metadata):
    """Write citation"""

    # Process the JSON data to generate a citeproc-py BibliographySource.
    item = write_citation_item(metadata)
    style_path = get_style_filepath(metadata.style)
    style = CitationStylesStyle(style_path, locale=metadata.locale)
    bib = CitationStylesBibliography(style, item, formatter.html)
    citation = Citation([CitationItem(metadata.id)])

    # workaround for the issue with the vancouver style and de locale
    try:
        bib.register(citation)
        return _clean_result(str(bib.bibliography()[0]))
    except Exception as e:
        print(e)
        return f"Error: citation not available for style {metadata.style} and locale {metadata.locale}."


def write_citation_item(metadata):
    """Write citation item"""
    if metadata.write_errors is not None:
        return None
    csl = json.loads(metadata.csl())

    # Remove keys that are not supported by citeproc-py.
    csl = py_.omit(csl, "copyright", "categories")

    return CiteProcJSON([csl])


def write_citation_list(metalist, **kwargs):
    """Write citation list"""
    if metalist is None:
        return None

    style = kwargs.get("style", "apa")
    locale = kwargs.get("locale", "en-US")
    style_path = get_style_filepath(style)
    style = CitationStylesStyle(style_path, locale=locale)  #

    def format_citation(index, item):
        bib = CitationStylesBibliography(style, item, formatter.html)
        _id = metalist.items[index].id
        citation = Citation([CitationItem(_id)])
        bib.register(citation)
        return _clean_result(str(bib.bibliography()[0]))

    citations = [write_citation_item(item) for item in metalist.items]
    bibliographies = [
        format_citation(index, item) for index, item in enumerate(citations)
    ]
    return "\n\n".join(bibliographies)


def _clean_result(text):
    """Remove double spaces, punctuation."""
    text = re.sub(r"\s\s+", " ", text)
    text = re.sub(r"\.\.+", ".", text)
    return text
